import argparse
import signal

import NDATools
from NDATools.BuildPackage import SubmissionPackage
from NDATools.Configuration import *
from NDATools.Submission import Submission
from NDATools.Utils import evaluate_yes_no_input, exit_client, get_request
from NDATools.Validation import Validation

logger = logging.getLogger(__name__)
def parse_args():
    parser = argparse.ArgumentParser(
        description='This application allows you to validate files and submit data into NDA. '
                    'You must enter a list of at least one file to be validated. '
                    'If your data contains manifest files, you must specify the location of the manifests. '
                    'If your data also includes associated files, you must enter a list of at least one directory '
                    'where the associated files are saved. Alternatively, if any of your data is stored in AWS, you must'
                    ' provide your account credentials, the AWS bucket, and a prefix, if it exists.  '
                    'Any files that are created while running the client (ie. results files) will be downloaded in '
                    'your home directory under NDAValidationResults. If your submission was interrupted in the middle'
                    ', you may resume your upload by entering a valid submission ID. ',
        usage='%(prog)s <file_list>')

    parser.add_argument('files', metavar='<file_list>', type=str, nargs='+', action='store',
                        help='Returns validation results for list of files')

    parser.add_argument('-l', '--listDir', metavar='<directory_list>', type=str, nargs='+', action='store',
                        help='Specifies the directories in which the associated files are files located.')

    parser.add_argument('-m', '--manifestPath', metavar='<arg>', type=str, nargs='+', action='store',
                        help='Specifies the directories in which the manifest files are located')

    parser.add_argument('-s3', '--s3Bucket', metavar='<arg>', type=str, action='store',
                        help='Specifies the s3 bucket in which the associated files are files located.')

    parser.add_argument('-pre', '--s3Prefix', metavar='<arg>', type=str, action='store',
                        help='Specifies the s3 prefix in which the associated files are files located.')

    parser.add_argument('-w', '--warning', action='store_true',
                        help='Returns validation warnings for list of files')

    parser.add_argument('-a', '--alternateEndpoint', metavar='<arg>', type=str, action='store',
                        help='An alternate upload location for the submission package')

    parser.add_argument('-b', '--buildPackage', action='store_true',
                        help='Flag whether to construct the submission package')

    parser.add_argument('-c', '--collectionID', metavar='<arg>', type=int, action='store',
                        help='The NDA collection ID')

    parser.add_argument('-d', '--description', metavar='<arg>', type=str, nargs='+', action='store',
                        help='The description of the submission')

    parser.add_argument('-p', '--password', help='Warning: Detected non-empty value for the -p/--password argument. '
                                                 'Support for this setting has been deprecated and will no longer be '
                                                 'used by this tool. Password storage is not recommended for security'
                                                 ' considerations')

    parser.add_argument('-t', '--title', metavar='<arg>', type=str, nargs='+', action='store',
                        help='The title of the submission')

    parser.add_argument('-u', '--username', metavar='<arg>', type=str, action='store',
                        help='NDA username')

    parser.add_argument('-ak', '--accessKey', metavar='<arg>', type=str, action='store',
                        help='AWS access key')

    parser.add_argument('-sk', '--secretKey', metavar='<arg>', type=str, action='store',
                        help='AWS secret key')

    parser.add_argument('-s', '--scope', metavar='<arg>', type=str, action='store',
                        help='Flag whether to validate using a custom scope. Must enter a custom scope')

    parser.add_argument('-rs', '--replace-submission', metavar='<arg>', type=str, action='store', default=0,
                        help='Use this arugment to replace a submission that has QA errors or that NDA staff has authorized manually to replace.')

    parser.add_argument('-r', '--resume', action='store_true',
                        help='Restart an in-progress submission, resuming from the last successful part in a multi-part'
                             'upload. Must enter a valid submission ID.')

    parser.add_argument('-v', '--validationAPI', metavar='<arg>', type=str, action='store',
                        help='URL of the validation tool API')

    parser.add_argument('-j', '--JSON', action='store_true',
                        help='Flag whether to additionally download validation results in JSON format.')

    parser.add_argument('-wt', '--workerThreads', metavar='<arg>', type=int, action='store',
                        help='Number of worker threads')

    parser.add_argument('-bc', '--batch', metavar='<arg>', type=int, action='store',
                        help='Batch size')

    parser.add_argument('--hideProgress', action='store_true', help='Hides upload/processing progress')

    parser.add_argument('--skipLocalAssocFileCheck', action='store_true', help='Not recommended UNLESS you have already'
                                                                               ' verified all paths for associated data files are correct')
    parser.add_argument('-f', '--force', action='store_true',
                        help='Ignores all warnings and continues without prompting for input from the user.')

    parser.add_argument('--validation-timeout', default=300, type=int, action='store', help='Timeout in seconds until the program errors out with an error. '
                                                                               'In most cases the default value of ''300'' seconds should be sufficient to validate submissions however it may'
                                                                               'be necessary to increase this value to a specific duration.')

    args = parser.parse_args()

    if args.password:
        print('Warning: Support for the password flag (-p, --password) has been removed from nda-tools due to security '
              'concerns and has been replaced with keyring.')
        args.__dict__.pop('password')

    return args


def configure(args):
    # create a new config file in user's home directory if one does not exist

    # always set password if --username flag is supplied, or if user is submitting data
    auth_req = True if args.buildPackage or args.resume or args.replace_submission or args.username else False

    if os.path.isfile(os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg')):
        config = ClientConfiguration(os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg'), args.username,
                                     args.accessKey, args.secretKey)
        config.read_user_credentials(auth_req)
    else:
        config = ClientConfiguration('clientscripts/config/settings.cfg', args.username, args.accessKey,
                                     args.secretKey)
        config.read_user_credentials(auth_req)
        config.make_config()

    if args.collectionID:
        config.collection_id = args.collectionID
    if args.alternateEndpoint:
        config.endpoint_title = args.alternateEndpoint
    if args.listDir:
        config.directory_list = args.listDir
    if args.manifestPath:
        config.manifest_path = args.manifestPath
    if args.s3Bucket:
        config.source_bucket = args.s3Bucket
    if args.s3Prefix:
        config.source_prefix = args.s3Prefix
    if args.validation_timeout:
        config.validation_timeout = args.validation_timeout
    if args.title:
        config.title = ' '.join(args.title)
    if args.description:
        config.description = ' '.join(args.description)
    if args.scope:
        config.scope = args.scope
    if args.validationAPI:
        config.validation_api = args.validationAPI[0]
    if args.JSON:
        config.JSON = True
    config.workerThreads = args.workerThreads
    config.hideProgress = args.hideProgress
    if args.skipLocalAssocFileCheck:
        config.skip_local_file_check = True
    if args.replace_submission:
        config.replace_submission = args.replace_submission
    config.force = True if args.force else False
    LoggingConfiguration.load_config(NDATools.NDA_TOOLS_VTCMD_LOGS_FOLDER)

    return config


class Status:
    UPLOADING = 'Uploading'
    SYSERROR = 'SystemError'


def resume_submission(submission_id, batch, config=None):
    submission = Submission( None, config, submission_id=submission_id, batch_size=batch, thread_num=config.workerThreads)
    submission.resume_submission()
    if submission.status != Status.UPLOADING:
        print_submission_complete_message(submission, False)


def validate_files(file_list, warnings, build_package, threads, config=None, pending_changes=None, original_uuids=None):
    validation = Validation(file_list, config=config, hide_progress=config.hideProgress, thread_num=threads,
                            allow_exit=True, pending_changes=pending_changes, original_uuids=original_uuids)
    logger.info('\nValidating files...')
    validation.validate()
    validation.output()

    if warnings:
        validation.get_warnings()
        logger.info('Warnings output to: {}'.format(validation.log_file))
    else:
        if validation.w:
            logger.info('\nNote: Your data has warnings. To save warnings, run again with -w argument.')
    logger.info('\nAll files have finished validating.')

    # Test if no files passed validation, exit
    if not any(map(lambda x: not validation.uuid_dict[x]['errors'], validation.uuid_dict)):
        logger.info('No files passed validation, please correct any errors and validate again.')
        sys.exit(1)
    # If some files passed validation, show files with and without errors
    else:
        logger.info('\nThe following files passed validation:')
        for uuid in validation.uuid_dict:
            if not validation.uuid_dict[uuid]['errors']:
                logger.info('UUID {}: {}'.format(uuid, validation.uuid_dict[uuid]['file']))
        if validation.e:
            logger.info('\nThese files contain errors:')
            for uuid in validation.uuid_dict:
                if validation.uuid_dict[uuid]['errors']:
                    logger.info('UUID {}: {}'.format(uuid, validation.uuid_dict[uuid]['file']))
    # If some files had errors, give option to submit just the files that passed
    if not hasattr(config, 'replace_submission'):
        # If some files had errors, give option to submit just the files that passed
        if build_package and validation.e and not config.force:
            proceed = evaluate_yes_no_input('Some files have errors, do you want to continue '
                                            'and submit ONLY the files that have passed validation?', 'n')

            if str(proceed).lower() == 'no':
                return
            elif str(proceed).lower() == 'y':
                validation.uuid = validation.verify_uuid()
    # We are replacing a submission
    else:
        if build_package and validation.e:
            logger.error('ERROR - At least some of the files failed validation. '
                  'All files must pass validation in order to edit submission {}. Please fix these errors and try again.'.format(
                config.replace_submission))
            sys.exit(1)
        elif build_package and validation.data_structures_with_missing_rows and not config.force:
            logger.warning('\nWARNING - Detected missing information in the following files: ')

            for tuple in validation.data_structures_with_missing_rows:
                logger.warning('\n{} - expected {} rows but found {}  '.format(tuple[0], tuple[1], tuple[2]))
            prompt = '\nIf you update your submission with these files, the missing data will be reflected in your data-expected numbers'
            prompt += '\nAre you sure you want to continue? <Yes/No>: '
            proceed = evaluate_yes_no_input(prompt, 'n')
            if str(proceed).lower() == 'n':
                exit_client(signal=signal.SIGTERM, message='')

    return validation.uuid, validation.associated_files_to_upload



def build_package(uuid, associated_files_to_upload, config, pending_changes=None, original_uuids=None):
    if not config.title:
        config.title = input('Enter title for dataset name:')
    if not config.description:
        config.description = input('Enter description for the dataset submission:')

    package = SubmissionPackage(uuid, associated_files_to_upload, config=config, allow_exit=True,
                                pending_changes=pending_changes, original_uuids=original_uuids)
    package.set_upload_destination(hide_input=False)
    directories = config.directory_list
    source_bucket = config.source_bucket
    source_prefix = config.source_prefix
    if associated_files_to_upload:
        logger.info('\nSearching for associated files...')
        package.file_search(directories, source_bucket, source_prefix, retry_allowed=True)
    logger.info('Building Package')
    package.build_package()
    logger.info('\n\nPackage Information:')
    logger.info('validation results: {}'.format(package.validation_results))
    logger.info('submission_package_uuid: {}'.format(package.submission_package_uuid))
    logger.info('created date: {}'.format(package.create_date))
    logger.info('expiration date: {}'.format(package.expiration_date))
    logger.info('\nPackage finished building.\n')

    logger.info('Downloading submission package.')
    package.download_package(hide_progress=config.hideProgress)
    logger.info('\nA copy of your submission package has been saved to: {}'.
          format(os.path.join(NDATools.NDA_TOOLS_SUB_PACKAGE_FOLDER, package.package_folder)))

    return [package.package_id, package.full_file_path]

def print_submission_complete_message(submission, replacement):
    if replacement:
        print('\nYou have successfully replaced submission {}.'.format(submission.submission_id))
    else:
        print('\nYou have successfully completed uploading files for submission {} with status: {}'.format
              (submission.submission_id, submission.status))

def submit_package(package_id, full_file_path, associated_files_to_upload, threads, batch,
                   config=None, original_submission_id=None):
    submission = Submission(package_id=package_id,
                            submission_id=original_submission_id,
                            full_file_path=full_file_path,
                            thread_num=threads,
                            batch_size=batch,
                            allow_exit=True,
                            config=config)
    logger.info('Requesting submission for package: {}'.format(submission.package_id))
    if original_submission_id:
        submission.replace_submission()
    else:
        submission.submit()
    # see commit comment for commit #d2f4dad
    # we need to trigger the GET /id endpoint to move the submission status to complete if necessary
    submission.check_status()
    if submission.submission_id:
        logger.info('Submission ID: {}'.format(str(submission.submission_id)))
    if associated_files_to_upload:
        logger.info('Preparing to upload associated files.')
        submission.upload_associated_files(hide_progress=config.hideProgress)
    if submission.status != Status.UPLOADING:
        print_submission_complete_message(submission, replacement=True if original_submission_id else False)


# sets self.pendingChanges and
def retrieve_replacement_submission_params(config, submission_id):
    # get submission-id
    api = type('', (), {})()
    api.config = config
    auth = requests.auth.HTTPBasicAuth(config.username, config.password)
    # check if the qa token provided is actually the latest or not
    try:
        response = get_request('/'.join([config.submission_api, submission_id, 'change-history']), auth=auth)
    except Exception as e:

        if e.response.status_code == 403:
            exit_client(signal=signal.SIGTERM,
                        message='You are not authorized to access submission {}. If you think this is a mistake, please contact NDA help desk'.format(
                            submission_id))
        else:
            exit_client(signal=signal.SIGTERM,
                        message='There was a General Error communicating with the NDA server. Please try again later')

    # TODO - check for 404 response

    # check to see if the submission was already replaced?
    if not response[0]['replacement_authorized']:
        if len(response) > 1 and response[1]['replacement_authorized']:
            message = '''Submission {} was already replaced by {} on {}.
If you need to make further edits to this submission, please reach out the the NDA help desk''' \
                .format(submission_id, response[0]['created_by'], response[0]['created_date'])
            exit_client(signal=signal.SIGTERM, message=message)
        else:
            exit_client(signal=signal.SIGTERM,
                        message='submission_id {} is not authorized to be replaced. Please contact the NDA help desk for approval to replace this submission'.format(
                            submission_id))

    response = get_request('/'.join([config.submission_api, submission_id]), auth=auth)
    if response is None:
        exit_client(signal=signal.SIGTERM,
                    message='There was a General Error communicating with the NDA server. Please try again later')

    submission_id = response['submission_id']
    config.title = response['dataset_title']
    config.description = response['dataset_description']
    config.collection_id = response['collection']['id']

    # get pending-changes for submission-id
    response = get_request('/'.join([config.submission_api, submission_id, 'pending-changes']), auth=auth);
    if response is None:
        exit_client(signal=signal.SIGTERM,
                    message='There was a General Error communicating with the NDA server. Please try again later')

    # get list of associated-files that have already been uplaoded for pending changes
    pending_changes = []
    original_submission_id = submission_id
    original_uuids = {uuid for uuid in response['validation_uuids']}
    for change in response['pendingChanges']:
        validation_uuids = change['validationUuids']
        associated_files = []
        manifest_files = []
        for uuid in validation_uuids:
            response = get_request('/'.join([config.validation_api, uuid]))
            associated_files.extend(response['associated_file_paths'])
            manifest_files.extend(manifest['localFileName'] for manifest in response['manifests'])
        change['associatedFiles'] = associated_files
        change['manifests'] = manifest_files
        pending_changes.append(change)

    return pending_changes, original_uuids, original_submission_id


def check_args(args):
    if args.replace_submission:
        if args.title or args.description or args.collectionID:
            message = 'Neither title, description nor collection_id arguments can be specified if' \
                      ' qa token is provided. Exiting...'
            logger.error(message)
            exit(1)


def main():
    # confirm most up to date version of nda-tools is installed
    args = parse_args()
    config = configure(args)

    pending_changes, original_uuids, original_submission_id = None, None, None
    check_args(args)
    if args.replace_submission:
        pending_changes, original_uuids, original_submission_id = retrieve_replacement_submission_params(config,
                                                                                                         args.replace_submission)

    if args.resume:
        submission_id = args.files[0]
        # Need to check to see if i need to update this step!
        resume_submission(submission_id, batch=args.batch, config=config)
    else:
        w = False
        bp = False
        if args.warning:
            w = True
        if args.buildPackage:
            bp = True
        validation_results = validate_files(args.files, w, bp, threads=args.workerThreads, config=config,
                                            pending_changes=pending_changes,
                                            original_uuids=original_uuids)
        if validation_results is not None:
            uuid = validation_results[0]
            associated_files_to_upload = validation_results[1]
            # If user requested to build a package
            if bp:
                package_results = build_package(uuid, associated_files_to_upload, config=config,
                                                pending_changes=pending_changes,
                                                original_uuids=original_uuids)
                package_id = package_results[0]
                full_file_path = package_results[1]
                submit_package(package_id=package_id, full_file_path=full_file_path,
                               associated_files_to_upload=associated_files_to_upload,
                               threads=args.workerThreads, batch=args.batch, config=config,
                               original_submission_id=original_submission_id)


if __name__ == "__main__":
    main()



from __future__ import absolute_import, with_statement

import getpass
import json
import logging
import logging.config
import platform
import shutil
import sys
import time

import keyring
import requests
import yaml
from requests import HTTPError

from NDATools import NDA_TOOLS_LOGGING_YML_FILE, Utils
from NDATools.Utils import exit_client, HttpErrorHandlingStrategy

if sys.version_info[0] < 3:
    import ConfigParser as configparser
    input = raw_input
else:
    import configparser
import os
from pkg_resources import resource_filename

logger = logging.getLogger(__name__)


class LoggingConfiguration:

    def __init__(self):
        pass

    @staticmethod
    def load_config(logs_directory):
        def make_config():
            file_path = os.path.join(os.path.expanduser('~'), '.NDATools')
            if not os.path.exists(file_path):
                os.makedirs(file_path)

            if not os.path.exists(NDA_TOOLS_LOGGING_YML_FILE):
                shutil.copyfile(resource_filename(__name__, 'clientscripts/config/logging.yml'),
                                NDA_TOOLS_LOGGING_YML_FILE)

        if not os.path.exists(NDA_TOOLS_LOGGING_YML_FILE):
            make_config()

        with open(NDA_TOOLS_LOGGING_YML_FILE, 'r') as stream:
            config = yaml.load(stream, Loader=yaml.FullLoader)
        log_file = os.path.join(logs_directory, "debug_log_{}.txt").format(time.strftime("%Y%m%dT%H%M%S"))
        config['handlers']['file']['filename']=log_file
        logging.config.dictConfig(config)

class ClientConfiguration:

    SERVICE_NAME = 'nda-tools'

    def __init__(self, settings_file, username=None, access_key=None, secret_key=None):
        self.config = configparser.ConfigParser()
        if settings_file == os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg'):
            self.config_location = settings_file
            self._check_and_fix_missing_options(self.config_location)
        else:
            self.config_location = resource_filename(__name__, settings_file)
        logger.info('Using configuration file from {}'.format(self.config_location))

        self.config.read(self.config_location)
        self.validation_api = self.config.get("Endpoints", "validation")
        self.submission_package_api = self.config.get("Endpoints", "submission_package")
        self.submission_api = self.config.get("Endpoints", "submission")
        self.validationtool_api = self.config.get("Endpoints", "validationtool")
        self.package_api = self.config.get("Endpoints", "package")
        self.datadictionary_api = self.config.get("Endpoints", "datadictionary")
        self.user_api = self.config.get("Endpoints", "user")
        self.aws_access_key = self.config.get("User", "access_key")
        self.aws_secret_key = self.config.get("User", "secret_key")
        self.aws_session_token = self.config.get('User', 'session_token')
        self.username = self.config.get("User", "username")

        self.check_deprecated_settings()

        if username:
            self.username = username
        elif self.username:
            logger.warning("-u/--username argument not provided. Using default value of '%s' which was saved in %s",
                           self.username, os.path.join(os.path.expanduser('~'), '.NDATools' + os.sep + 'settings.cfg'))

        if access_key:
            self.aws_access_key = access_key
        if secret_key:
            self.aws_secret_key = secret_key
        self.collection_id = None
        self.endpoint_title = None
        self.scope = None
        self.directory_list = None
        self.manifest_path = None
        self.source_bucket = None
        self.source_prefix = None
        self.title = None
        self.description = None
        self.JSON = False
        self.hideProgress = False
        self.skip_local_file_check = False
        logger.info('proceeding as nda user: {}'.format(self.username))

    def _check_and_fix_missing_options(self, config_location):
        default_config = configparser.ConfigParser()
        default_file_path = resource_filename(__name__, 'clientscripts/config/settings.cfg')
        default_config.read(default_file_path)
        default_settings = dict(default_config._sections)
        self.config.read(config_location)
        user_settings = dict(self.config._sections)

        for section in default_settings:
            for option in default_settings[section]:
                if option not in user_settings[section]:
                    logger.debug('[{}][{}] is missing'.format(section, option))
                    with open(config_location, 'w') as configfile:
                        self.config.set(section, option, default_settings[section][option])
                        self.config.write(configfile)

    def make_config(self):
        file_path = os.path.join(os.path.expanduser('~'), '.NDATools')
        if not os.path.exists(file_path):
            os.makedirs(file_path)
        config_path = os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg')

        copy_config = configparser.ConfigParser()

        copy_config.add_section("Endpoints")
        copy_config.set("Endpoints", "user", self.user_api)
        copy_config.set("Endpoints", "package", self.package_api)
        copy_config.set("Endpoints", "validation", self.validation_api)
        copy_config.set("Endpoints", "submission_package", self.submission_package_api)
        copy_config.set("Endpoints", "submission", self.submission_api)
        copy_config.set("Endpoints", "validationtool", self.validationtool_api)
        copy_config.set("Endpoints", "datadictionary", self.datadictionary_api)

        copy_config.add_section("User")
        copy_config.set("User", "username", self.username)
        copy_config.set("User", "access_key", self.aws_access_key)
        copy_config.set("User", "secret_key", self.aws_secret_key)
        copy_config.set("User", "session_token", self.aws_session_token)

        with open(config_path, 'w') as configfile:
            copy_config.write(configfile)

    def read_user_credentials(self, auth_req=True):

        self.password = None
        if auth_req:
            while True:
                try:
                    while not self.username:
                        self.username = input('Enter your NIMH Data Archives username:')
                        with open(self.config_location, 'w') as configfile:
                            self.config.set('User', 'username', self.username)
                            self.config.write(configfile)
                    self.password = keyring.get_password(self.SERVICE_NAME, self.username)
                    while not self.password:
                        self.password = getpass.getpass('Enter your NIMH Data Archives password:')
                    while not self.is_valid_nda_credentials():
                        logger.error("The password that was entered for user '%s' is invalid ...", self.username)
                        logger.error('If your username was previously entered incorrectly, you may update it in your '
                                     'settings.cfg located at \n{}'.format(self.config_location))
                        self.password = getpass.getpass('Enter your NIMH Data Archives password:')
                        while self.password == '':
                            self.password = getpass.getpass('Enter your NIMH Data Archives password:')
                    else:
                        keyring.set_password(self.SERVICE_NAME, self.username, self.password)
                        break
                except RuntimeError as e:
                    if platform.system() == 'Linux':
                        print('If there is no backend set up for keyring, you may try\n'
                              'pip install secretstorage --upgrade keyrings.alt')
                    raise e

        # Only ask for access-key/secret-key when needed (which is only when a user is creating a submission
        # and the files are stored in an s3 bucket.
        if self.source_bucket:
            self.read_aws_credentials()

    def read_aws_credentials(self):
        if not self.aws_access_key:
            self.aws_access_key = getpass.getpass("Enter your aws_access_key (must have read access to the %s bucket): " % self.source_bucket)

        if not self.aws_secret_key:
            self.aws_secret_key = getpass.getpass('Enter your aws_secret_key:')

    def check_deprecated_settings(self):
        if self.config.has_option('User', 'password') and self.config.get("User", "password"):
            print('Warning: Detected non-empty value for "password" in settings.cfg. Support for this setting has been '
                  'deprecated and will no longer be used by this tool. Password storage is not recommended for security'
                  ' considerations')

    def is_valid_nda_credentials(self):
        try:
            # will raise HTTP error 401 if invalid creds
            tmp = Utils.get_request(self.user_api, headers={'content-type': 'application/json'},
                                    auth=requests.auth.HTTPBasicAuth(self.username, self.password),
                                    error_handler=HttpErrorHandlingStrategy.reraise_status)
            return True
        except HTTPError as e:
            if e.response.status_code == 401:
                if 'locked' in e.response.text:
                    # user account is locked
                    tmp = json.loads(e.response.text)
                    logger.error('\nError: %s', tmp['message'])
                    logger.error('\nPlease contact NDAHelp@mail.nih.gov for help in resolving this error')
                    exit_client()
                else:
                    return False
            else:
                logger.error('\nSystemError while checking credentials for user %s', self.username)
                logger.error('\nPlease contact NDAHelp@mail.nih.gov for help in resolving this error')
                exit_client()



from __future__ import print_function

import json
import os
import sys

import requests

import NDATools

__version__ = '0.2.24'
pypi_version = None
version_checked = False


def check_version():
    try:
        from packaging.version import parse
    except ImportError:
        from pip._vendor.packaging.version import parse
    # use https://test.pypi.org/pypi/{package}/json on test/release branches, use https://pypi.org on master
    # Print Python 2 support dropping warning
    if sys.version_info < (3, 0):
        print()
        print('''WARNING - Detected Python version 2. Support for Python 2 is being removed from nda-tools. It is recommended to upgrade to the latest version of Python 
          before using the latest features in the downloadcmd''')
    print('Running NDATools Version {}'.format(NDATools.__version__))
    if parse(NDATools.__version__).is_devrelease:
        return
    url_pattern = 'https://pypi.org/pypi/{package}/json'
    package = 'nda-tools'
    """Return version of package on pypi.python.org using json."""
    req = requests.get(url_pattern.format(package=package))
    version = parse('0')
    if req.status_code == requests.codes.ok:
        j = json.loads(req.text)
        releases = j.get('releases', [])
        for release in releases:
            ver = parse(release)
            if not ver.is_prerelease:
                version = max(version, ver)
    NDATools.pypi_version = str(version)

    if parse(NDATools.__version__) < parse(NDATools.pypi_version):
        print("Your version of nda-tools is out of date. Please upgrade to the latest version ({}) from PyPi or GitHub and "
              "try again. \n\tTo upgrade using pip, run: \r\npip install nda-tools=={}".format(NDATools.pypi_version, NDATools.pypi_version))
        sys.exit(1)

    NDATools.version_checked = True

if not NDATools.version_checked:
    check_version()

# init folder structure for program runtime files
NDA_ORGINIZATION_ROOT_FOLDER = os.path.join(os.path.expanduser('~'), 'NDA')
if not os.path.exists(NDA_ORGINIZATION_ROOT_FOLDER):
    os.mkdir(NDA_ORGINIZATION_ROOT_FOLDER)

NDA_TOOLS_ROOT_FOLDER = os.path.join(NDA_ORGINIZATION_ROOT_FOLDER, 'nda-tools')
if not os.path.exists(NDA_TOOLS_ROOT_FOLDER):
    os.mkdir(NDA_TOOLS_ROOT_FOLDER)

NDA_TOOLS_VTCMD_FOLDER = os.path.join(NDA_TOOLS_ROOT_FOLDER, 'vtcmd')
if not os.path.exists(NDA_TOOLS_VTCMD_FOLDER):
    os.mkdir(NDA_TOOLS_VTCMD_FOLDER)

NDA_TOOLS_DOWNLOADCMD_FOLDER = os.path.join(NDA_TOOLS_ROOT_FOLDER, 'downloadcmd')
if not os.path.exists(NDA_TOOLS_DOWNLOADCMD_FOLDER):
    os.mkdir(NDA_TOOLS_DOWNLOADCMD_FOLDER)


NDA_TOOLS_DOWNLOADS_FOLDER = os.path.join(NDA_TOOLS_DOWNLOADCMD_FOLDER, 'packages')
if not os.path.exists(NDA_TOOLS_DOWNLOADS_FOLDER):
    os.mkdir(NDA_TOOLS_DOWNLOADS_FOLDER)

NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER = os.path.join(NDA_TOOLS_DOWNLOADCMD_FOLDER, 'logs')
if not os.path.exists(NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER):
    os.mkdir(NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER)

NDA_TOOLS_VTCMD_LOGS_FOLDER = os.path.join(NDA_TOOLS_VTCMD_FOLDER, 'logs')
if not os.path.exists(NDA_TOOLS_VTCMD_LOGS_FOLDER):
    os.mkdir(NDA_TOOLS_VTCMD_LOGS_FOLDER)

NDA_TOOLS_VAL_FOLDER = os.path.join(NDA_TOOLS_VTCMD_FOLDER, 'validation_results')
if not os.path.exists(NDA_TOOLS_VAL_FOLDER):
    os.mkdir(NDA_TOOLS_VAL_FOLDER)

NDA_TOOLS_SUB_PACKAGE_FOLDER = os.path.join(NDA_TOOLS_VTCMD_FOLDER, 'submission_package')
if not os.path.exists(NDA_TOOLS_SUB_PACKAGE_FOLDER):
    os.mkdir(NDA_TOOLS_SUB_PACKAGE_FOLDER)

NDA_TOOLS_DEFAULT_LOG_FORMAT='%(asctime)s:%(levelname)s:%(message)s'

NDA_TOOLS_LOGGING_YML_FILE = os.path.join(os.path.expanduser('~'), '.NDATools/logging.yml')

os.environ['PYTHONWARNINGS']='ignore' # MAC users sometimes see output from python warnings module. Suppress these msgs



from __future__ import absolute_import, with_statement

import datetime
import functools
import json
import logging
import os
import random
import re
import signal
import sys
import threading
import time
import traceback

import boto3
import requests
from requests.adapters import HTTPAdapter
import requests.packages.urllib3.util

IS_PY2 = sys.version_info < (3, 0)

if IS_PY2:
    from io import open
    from urlparse import urlparse
else:
    from urllib.parse import urlparse

logger = logging.getLogger(__name__)


if sys.version_info[0] < 3:
    input = raw_input
else:
    pass


class Protocol(object):
    CSV = "csv"
    XML = "xml"
    JSON = "json"

    @staticmethod
    def get_protocol(cls):
        return cls.JSON

class DeserializeHandler():

    @staticmethod
    def none(r):
        return r

    @staticmethod
    def convert_json(r):
        return json.loads(r.text)

class HttpErrorHandlingStrategy():
    # error handling implementation methods. Each method takes a response object
    @staticmethod
    def ignore(r):
        pass


    @staticmethod
    def print_and_exit(r):
        # handle json and plain-text errors
        message = None
        try:
            if 'content-type' in r.headers and 'json' in r.headers['Content-Type']:
                message = r.json()['message']
        except (ValueError, json.JSONDecodeError):
            message = r.text

        if r.status_code == 401:
            # provide default message if one doesnt already exist
            message = message or 'The NDA username or password is not recognized.'
        else:
            message = '\nAn unexpected error was encountered and the program could not continue. Error message from service was: \n%s' % message
        logger.error(message)
        exit_client()

    @staticmethod
    def reraise_status(response):
        response.raise_for_status()

def exit_client(signal=signal.SIGTERM, frame=None, message=None):
    for t in threading.enumerate():
        try:
            t.shutdown_flag.set()
        except AttributeError as e:
            continue
    if message:
        logger.info('\n\n{}'.format(message))
    else:
        logger.info('\n\nExit signal received, shutting down...')
    os._exit(1)


def parse_local_files(directory_list, no_match, full_file_path, no_read_access, skip_local_file_check):
    """
    Iterates through associated files generate a dictionary of full filepaths and file sizes.

    :param directory_list: List of directories
    :param no_match: Stores list of invalid paths
    :param full_file_path: Dictionary of tuples that store full filepath and file size
    :param no_read_access: List of files that user does not have access to
    :return: Modifies references to no_match, full_file_path, no_read_access
    """
    files_to_match = len(no_match)
    logger.debug('Starting local directory search for {} files'.format(str(files_to_match)))
    for file in no_match[:]:
        file_key = sanitize_file_path(file)
        for d in directory_list:
            if skip_local_file_check:
                file_name = os.path.join(d, file)
                try:
                    full_file_path[file_key] = (file_name, os.path.getsize(file_name))
                    no_match.remove(file)
                except (OSError, IOError) as err:
                    if err.errno == 13:
                        logger.info('Permission Denied: {}'.format(file_name))
                    continue
                break
            else:
                if os.path.isfile(file):
                    file_name = file
                elif os.path.isfile(os.path.join(d, file)):
                    file_name = os.path.join(d, file)
                else:
                    continue
                if not check_read_permissions(file_name):
                    no_read_access.add(file_name)
                full_file_path[file_key] = (file_name, os.path.getsize(file_name))
                no_match.remove(file)
                sys.stdout.write(
                    '\rFound {} files out of {}\r'.format(str(files_to_match - len(no_match)), str(files_to_match)))
                sys.stdout.flush()
                break
    sys.stdout.write('{}\r'.format(' '*64)) # clear out 'Found {} files ...' message and reset cursor
    sys.stdout.flush()
    logger.debug(
        'Local directory search complete, found {} files out of {}'.format(str(files_to_match - len(no_match)),
                                                                           str(files_to_match)))


def sanitize_file_path(file):
    """
    Replaces backslashes with forward slashes and removes leading / or drive:/.

    :param file: Relative or absolute filepath
    :return: Relative or absolute filepath with leading / or drive:/ removed
    """
    # Sanitize all backslashes (\) with forward slashes (/)
    file_key = file.replace('\\', '/').replace('//', '/')
    # If Mac/Linux full path
    if re.search(r'^/.+$', file):
        file_key = file.split('/', 1)[1]
    # If Windows full path
    elif re.search(r'^\D:/.+$', file_key):
        file_key = file_key.split(':/', 1)[1]
    return file_key


def check_read_permissions(file):
    try:
        open(file)
        return True
    except (OSError, IOError) as err:
        if err.errno == 13:
            logger.info('Permission Denied: {}'.format(file))
    return False


def evaluate_yes_no_input(message, default_input=None):
    while True:
        default_print = ' (Y/n)' if default_input.upper() == 'Y' else  ' (y/N)' if default_input.upper() == 'N' else ''
        user_input = input('{}{}'.format(message, default_print)) or default_input
        if str(user_input).upper() in ['Y', 'N']:
            return user_input.lower()
        else:
            print('Input not recognized.')


def get_error():
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = traceback.TracebackException(
        exc_type, exc_value, exc_tb,
    )
    ex = ''.join(tbe.format_exception_only())
    return 'Error: {}'.format(ex)


def get_traceback():
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = traceback.TracebackException(
        exc_type, exc_value, exc_tb,
    )
    tb = ''.join(tbe.format())
    return tb


# return bucket and key for url (handles http and s3 protocol)
def deconstruct_s3_url(url):
    tmp = urlparse(url)
    if tmp.scheme == 's3':
        bucket = tmp.netloc
        path = tmp.path.lstrip('/')
    elif tmp.scheme == 'https':
        # presigned urls are either https://bucketname.s3.amazonaws.com/key... or
        # https://s3.amazonaws.com/bucket/path
        if tmp.hostname == 's3.amazonaws.com':
            bucket = tmp.path.split('/')[1]
            path = '/'.join(tmp.path.split('/')[2:])
            path = '/' + path
        else:
            bucket = tmp.hostname.replace('.s3.amazonaws.com', '')
            path = tmp.path
    else:
        raise Exception('Invalid URL passed to deconstruct_s3_url method: {}'.format(url))

    return bucket, path.lstrip('/')


# converts . and .. and ~ in file-paths. (as well as variable names like %HOME%
def convert_to_abs_path(file_name):
    return os.path.abspath(os.path.expanduser(os.path.expandvars(file_name)))


def human_size(bytes, units=[' bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB']):
    """ Returns a human readable string representation of bytes """
    return str(round(bytes, 2)) + units[0] if bytes < 1024 else human_size(bytes / 1024, units[1:])

def retry_connection_errors(func):
    @functools.wraps(func)
    def _retry(*args, **kwargs):
        tmp = None
        for i in range(10):
            try:
                tmp = func(*args, **kwargs)
                return tmp
            except requests.exceptions.ConnectionError as e:
                if i == 9:
                    raise e
                time.sleep(random.randint(10, 30))
    return _retry

def is_json(test):
    try:
        json.dumps(test)
        return True
    except:
        return False

@retry_connection_errors
def _send_prepared_request(prepped, timeout=150, deserialize_handler=DeserializeHandler.convert_json, error_handler=HttpErrorHandlingStrategy.print_and_exit):
    with requests.Session() as session:
        logger.debug('{} {} @ {}'.format(prepped.method , prepped.url, datetime.datetime.now()))
        session.mount(prepped.url, HTTPAdapter(max_retries=10))
        tmp = session.send(prepped, timeout=timeout)
        logger.debug('{} {} (elapsed = {})- STATUS {}'.format(prepped.method, prepped.url, tmp.elapsed, tmp.status_code))
        if not tmp.ok:
            error_handler(tmp)
    return deserialize_handler(tmp)


def get_request(url, headers={}, auth=None, timeout=150, deserialize_handler=DeserializeHandler.convert_json, error_handler=HttpErrorHandlingStrategy.print_and_exit):
    req = requests.Request('GET', url, auth=auth, headers=headers)
    return _send_prepared_request(req.prepare(), timeout=timeout, deserialize_handler=deserialize_handler, error_handler=error_handler)

def post_request(url, payload=None, headers={}, auth=None, timeout=150, deserialize_handler=DeserializeHandler.convert_json, error_handler=HttpErrorHandlingStrategy.print_and_exit):
    data_param, headers = get_data_and_header_params(payload, headers)
    req = requests.Request('POST',  url, auth=auth,  headers=headers, **data_param)
    return _send_prepared_request(req.prepare(), timeout=timeout, deserialize_handler=deserialize_handler, error_handler=error_handler)

def put_request(url, payload=None, headers={}, auth=None, timeout=150, deserialize_handler=DeserializeHandler.convert_json, error_handler=HttpErrorHandlingStrategy.print_and_exit):
    data_param, headers = get_data_and_header_params(payload, headers)
    req = requests.Request('PUT',  url, auth=auth, headers=headers, **data_param)
    return _send_prepared_request(req.prepare(), timeout=timeout, deserialize_handler=deserialize_handler, error_handler=error_handler)

def get_data_and_header_params(payload, headers):
    data_param = {}
    if 'content-type' not in headers:
        if isinstance(payload, dict) or isinstance(payload, list):
            # setting the json arg will automatically add the content-header
            # https://requests.readthedocs.io/en/latest/user/quickstart/#more-complicated-post-requests
            data_param ={'json': payload}
        else:
            data_param = {'data': payload}
            if isinstance(payload, str) and is_json(payload) and 'content-type' not in headers:
                headers['content-type'] = 'application/json'
    else:
        data_param = {'data': payload}
    return data_param, headers

def get_s3_client_with_config(aws_access_key, aws_secret_key, aws_session_token):
    return boto3.session.Session(aws_access_key_id=aws_access_key,
                                 aws_secret_access_key=aws_secret_key,
                                 aws_session_token=aws_session_token,
                                 region_name='us-east-1').client('s3')


def get_s3_resource(aws_access_key, aws_secret_key, aws_session_token, s3_config):
    return boto3.Session(aws_access_key_id=aws_access_key,
                         aws_secret_access_key=aws_secret_key,
                         aws_session_token=aws_session_token).resource('s3', config=s3_config)



from __future__ import absolute_import, with_statement

import csv
import multiprocessing
import sys

from tqdm import tqdm

import NDATools

if sys.version_info[0] < 3:
    import Queue as queue

    input = raw_input
else:
    import queue
from NDATools.Configuration import *
from NDATools.Utils import *
import threading
import signal

logger = logging.getLogger(__name__)
class Validation:
    def __init__(self, file_list, config, hide_progress, allow_exit=False, thread_num=None,
                 pending_changes=None, original_uuids=None):
        self.config = config
        self.hide_progress = hide_progress
        self.api = self.config.validation_api.strip('/')
        self.thread_num = max([1, multiprocessing.cpu_count() - 1])
        if thread_num:
            self.thread_num = thread_num
        self.scope = self.config.scope
        self.api_scope = self.api
        if self.scope is not None:
            self.api_scope = "".join([self.api, '/?scope={}'.format(self.scope)])

        self.file_queue = queue.Queue()
        self.result_queue = queue.Queue()
        self.file_list = file_list
        self.directory_list = self.config.directory_list
        self.associated_files_to_upload = set()
        self.uuid = []
        self.uuid_dict = {}
        self.responses = []
        self.date = time.strftime("%Y%m%dT%H%M%S")
        self.e = False
        self.w = False
        self.manifest_data_files = []
        self.manifest_path = self.config.manifest_path
        if self.config.JSON:
            self.log_file = os.path.join(NDATools.NDA_TOOLS_VAL_FOLDER, 'validation_results_{}.json'.format(self.date))
        else:
            self.log_file = os.path.join(NDATools.NDA_TOOLS_VAL_FOLDER, 'validation_results_{}.csv'.format(self.date))

        self.field_names = ['FILE', 'ID', 'STATUS', 'EXPIRATION_DATE', 'ERRORS', 'COLUMN', 'MESSAGE', 'RECORD']
        self.validation_progress = None
        self.pending_changes = pending_changes
        self.exit = allow_exit
        self.original_uuids = original_uuids
        self.data_structures_with_missing_rows = None
        if self.config.password:
            self.auth = requests.auth.HTTPBasicAuth(self.config.username, self.config.password)
        else:
            self.auth = None

    """
    Validates a list of csv files and saves a new csv file with the results in the Home directory.
    If data includes associated files and user has not entered any directories where they are saved, user is prompted to 
    enter a list of directories only if they have also indicated to build a package for submission.
    """

    def validate(self):
        if not self.hide_progress:
            self.validation_progress = tqdm(total=len(self.file_list), position=0, ascii=os.name == 'nt')
        workers = []
        for x in range(self.thread_num):
            worker = Validation.ValidationTask(self.file_queue, self.result_queue, self.api, self.api_scope,
                                               self.responses, self.validation_progress, self.exit,
                                               self.config.validation_timeout, self.auth)

            workers.append(worker)
            worker.daemon = True
            worker.start()
        for file in self.file_list:
            self.file_queue.put(file)
        self.file_queue.put("STOP")
        while any(map(lambda x: x.is_alive(), workers)):
            time.sleep(2)
        if not self.hide_progress:
            self.validation_progress.close()
        while True:
            try:
                (response, file) = self.result_queue.get()
                self.result_queue.task_done()
                self.responses.append((response, file))
                if self.result_queue.empty():
                    break
            except queue.Empty:
                break
        for (response, file) in self.responses:
            if response['status'] == Status.SYSERROR:
                self.e = True
                error = " ".join([response['status'], response['errors']['system'][0]['message']])
                raise Exception(error)
            elif response['errors'] != {}:
                self.e = True
            if response['associated_file_paths'] and response['errors'] == {}:
                self.associated_files_to_upload.update(set(response['associated_file_paths']))
            if response['warnings'] != {}:
                self.w = True
            if response['status'] == Status.PENDING_MANIFEST and response['errors'] == {}:
                self.process_manifests(response)
                response = get_request("/".join([self.api, response['id']]), auth=self.auth)
                self.associated_files_to_upload.update(set(response['associated_file_paths']))

            self.uuid.append(response['id'])
            self.uuid_dict[response['id']] = {
                'file': file, 'errors': response['errors'] != {},
                'short_name': response['short_name'],
                'associated_file_paths': set(response['associated_file_paths']),
                'rows': response['rows'],
                'manifests': {manifest['localFileName'] for manifest in response['manifests']}
            }

        if self.pending_changes:
            # remove the associated_files that have already been uplaoded
            structure_to_new_associated_files = {}
            unrecognized_ds = set()
            for uuid in self.uuid_dict:
                short_name = self.uuid_dict[uuid]['short_name']
                structure_to_new_associated_files[short_name] = set()
            for uuid in self.uuid_dict:
                short_name = self.uuid_dict[uuid]['short_name']
                structure_to_new_associated_files[short_name].update(
                    {file for file in self.uuid_dict[uuid]['associated_file_paths']})

            files_to_upload = set()
            for data_structure in structure_to_new_associated_files:
                expected_change_for_data_structure = next(
                    filter(lambda pending_change: pending_change['shortName'] == data_structure, self.pending_changes),
                    None)
                if expected_change_for_data_structure is not None:
                    original_associated_file_set = {associated_file for associated_file in
                                                    expected_change_for_data_structure['associatedFiles']}
                    for new_asssociated_file in structure_to_new_associated_files[data_structure]:
                        if new_asssociated_file not in original_associated_file_set:
                            files_to_upload.update({new_asssociated_file})
                else:
                    unrecognized_ds.update({data_structure})


            self.associated_files_to_upload = set(files_to_upload)
            if len(files_to_upload) > 0:
                logger.info ('\nDetected {} new files that need to be uploaded for submission.\n'.format(len(files_to_upload)))
            else:
                logger.info('\nDetected that all associated files have been previously uploaded from previous submission\n')

            # Find datastructures with missing data
            # create a map of short_name to num-rows for files that the user uploaded
            structure_to_new_row_count = {}
            for uuid in self.uuid_dict:
                short_name = self.uuid_dict[uuid]['short_name']
                structure_to_new_row_count[short_name] = 0
            for uuid in self.uuid_dict:
                short_name = self.uuid_dict[uuid]['short_name']
                structure_to_new_row_count[short_name] += self.uuid_dict[uuid]['rows']

            data_structures_with_missing_rows = []
            for data_structure in structure_to_new_row_count:
                expected_change_for_data_structure = next(
                    filter(lambda pending_change: pending_change['shortName'] == data_structure, self.pending_changes),
                    None)
                if expected_change_for_data_structure is not None:
                    if structure_to_new_row_count[data_structure] < expected_change_for_data_structure['rows']:
                        data_structures_with_missing_rows.append((data_structure,
                                                                  expected_change_for_data_structure['rows'],
                                                                  structure_to_new_row_count[data_structure]))
                else:
                    unrecognized_ds.update({data_structure})

            # update list of validation-uuids to be used during the packaging step
            new_uuids, unrecognized_ds = self.generate_uuids_for_qa_workflow(unrecognized_ds)

            if unrecognized_ds:
                message = 'ERROR - The following datastructures were not included in the original submission and therefore cannot be included in the replacement submission: '
                message += "\r\n" + "\r\n".join(unrecognized_ds)
                exit_client(signal=signal.SIGTERM, message=message)
            else:
                self.data_structures_with_missing_rows = data_structures_with_missing_rows
                self.uuid = new_uuids

    def get_existing_manifests(self):
        if not self.pending_changes:
            return set()
        # create a set of manifest files from the set of pending changes to enforce no duplicates
        # convert back into a list because the process_manifests method expects a list
        return {manifest for manifests in list(map(lambda change: change['manifests'], self.pending_changes)) for manifest
             in manifests}

    def output(self):
        encountered_system_error=False
        if self.config.JSON:
            json_data = dict(Results=[])
            for (response, file) in self.responses:
                if response['status'] == Status.SYSERROR:
                    encountered_system_error=True
                file_name = self.uuid_dict[response['id']]['file']
                json_data['Results'].append({
                    'File': file_name,
                    'ID': response['id'],
                    'Status': response['status'],
                    'Expiration Date': response['expiration_date'],
                    'Errors': response['errors']
                })
            with open(self.log_file, 'w') as outfile:
                json.dump(json_data, outfile)

        else:
            if sys.version_info[0] < 3:
                csvfile = open(self.log_file, 'wb')
            else:
                csvfile = open(self.log_file, 'w', newline='')
            writer = csv.DictWriter(csvfile, fieldnames=self.field_names)
            writer.writeheader()
            for (response, file) in self.responses:
                if response['status'] == Status.SYSERROR:
                    encountered_system_error=True

                file_name = self.uuid_dict[response['id']]['file']

                if response['errors'] == {}:
                    writer.writerow(
                        {'FILE': file_name, 'ID': response['id'], 'STATUS': response['status'],
                         'EXPIRATION_DATE': response['expiration_date'], 'ERRORS': 'None', 'COLUMN': 'None',
                         'MESSAGE': 'None', 'RECORD': 'None'})
                else:
                    for error, value in response['errors'].items():
                        if error == 'system':
                            encountered_system_error = True
                        for v in value:
                            try:
                                column = v['columnName'] if 'columnName' in v else None
                                message = v['message']
                                record = v['recordNumber'] if 'recordNumber' in v else None
                                writer.writerow(
                                    {'FILE': file_name, 'ID': response['id'], 'STATUS': response['status'],
                                     'EXPIRATION_DATE': response['expiration_date'], 'ERRORS': error, 'COLUMN': column,
                                     'MESSAGE': message, 'RECORD': record})
                            except KeyError:
                                logger.error('Unrecognized result from validation service - {}. '.format(json.dumps(response)))
                                logger.error('\nPlease contact NDAHelp@mail.nih.gov for help in resolving this error')
                                exit_client()
        csvfile.close()
        logger.info('Validation report output to: {}'.format(self.log_file))
        if encountered_system_error:
            logger.error('Unexpected error occurred while validating one or more of the csv files')
            logger.error('\nPlease email NDAHelp@mail.nih.gov for help in resolving this error and include {} as an attachment to help us resolve the issue'.format(self.log_file))
            exit_client()

    def get_warnings(self):
        if self.config.JSON:
            json_data = dict(Results=[])
            for (response, file) in self.responses:
                file_name = self.uuid_dict[response['id']]['file']
                json_data['Results'].append({
                    'File': file_name,
                    'ID': response['id'],
                    'Status': response['status'],
                    'Expiration Date': response['expiration_date'],
                    'Warnings': response['warnings']
                })
                new_path = ''.join([NDATools.NDA_TOOLS_VAL_FOLDER, '/validation_warnings_', self.date, '.json'])
                with open(new_path, 'w') as outfile:
                    json.dump(json_data, outfile)
        else:
            new_path = ''.join([NDATools.NDA_TOOLS_VAL_FOLDER, '/validation_warnings_', self.date, '.csv'])
            if sys.version_info[0] < 3:
                csvfile = open(new_path, 'wb')
            else:
                csvfile = open(new_path, 'w', newline='')
            fieldnames = ['FILE', 'ID', 'STATUS', 'EXPIRATION_DATE', 'WARNINGS', 'MESSAGE', 'COUNT']
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()
            for (response, file) in self.responses:
                file_name = self.uuid_dict[response['id']]['file']
                if response['warnings'] == {}:
                    writer.writerow(
                        {'FILE': file_name, 'ID': response['id'], 'STATUS': response['status'],
                         'EXPIRATION_DATE': response['expiration_date'], 'WARNINGS': 'None', 'MESSAGE': 'None',
                         'COUNT': '0'})
                else:
                    for warning, value in response['warnings'].items():
                        m = {}
                        warning_list = []
                        for v in value:
                            count = 1
                            if v not in warning_list:
                                warning_list.append(v)
                                message = v['message']
                                m[message] = count
                            else:
                                count = m[message] + 1
                                m[message] = count
                        for x in m:
                            writer.writerow(
                                {'FILE': file_name,
                                 'ID': response['id'],
                                 'STATUS': response['status'],
                                 'EXPIRATION_DATE': response['expiration_date'],
                                 'WARNINGS': warning,
                                 'MESSAGE': x,
                                 'COUNT': m[x]})
            csvfile.close()

    def verify_uuid(self):
        uuid_list = []
        for uuid in self.uuid_dict:
            if not self.uuid_dict[uuid]['errors']:
                uuid_list.append(uuid)

        for response in self.manifest_data_files:
            if response['id'] not in uuid_list:
                self.manifest_data_files.remove(response)
        if self.original_uuids:
            uuid_list = self.generate_uuids_for_qa_workflow()[0]
        return uuid_list

    def generate_uuids_for_qa_workflow(self, unrecognized_ds=set()):
        unrecognized_structures = set(unrecognized_ds)
        new_uuids = set(self.original_uuids)
        val_by_short_name = {}
        for uuid in self.uuid_dict:
            short_name = self.uuid_dict[uuid]['short_name']
            val_by_short_name[short_name] = set()
        for uuid in self.uuid_dict:
            short_name = self.uuid_dict[uuid]['short_name']
            val_by_short_name[short_name].update({uuid})
        for short_name in val_by_short_name:
            # find the pending change with the same short name
            matching_change = {}
            for change in self.pending_changes:
                if change['shortName'] == short_name:
                    matching_change = change
            if not matching_change:
                unrecognized_structures.add(short_name)
            else:
                # prevValidationUuids is the set of validation-uuids on the pending changes resource
                new_uuids = new_uuids.difference(set(matching_change['validationUuids']))
                new_uuids.update({res for res in val_by_short_name[short_name]})

        return list(new_uuids), unrecognized_structures

    def process_manifests(self, r, validation_results=None, yes_manifest=set(), bulk_upload=False):
        if not self.manifest_path:
            if not self.exit:
                error = 'Missing Manifest File: You must include the path to your manifests files'
                raise Exception(error)
            else:
                manifest_path_input = input("\nYour data contains manifest files. Please enter a list of "
                                            "the complete paths to the folder where your manifests are located,"
                                            "separated by a space:")
                self.manifest_path = manifest_path_input.split(' ')
        no_manifest = set()

        if validation_results:
            self.validation_result = validation_results
        else:
            self.validation_result = Validation.ValidationManifestResult(r, self)
        if not bulk_upload:
            files_to_upload = len(self.validation_result.manifests) - len(yes_manifest)
            uploaded_count = 0
            for validation_manifest in self.validation_result.manifests:
                for m in self.manifest_path:
                    if validation_manifest.local_file_name not in yes_manifest:
                        try:
                            manifest_path = os.path.join(m, validation_manifest.local_file_name)
                            validation_manifest.upload_manifest(open(manifest_path, 'rb'))
                            yes_manifest.add(validation_manifest.local_file_name)
                            no_manifest.discard(validation_manifest.local_file_name)
                            uploaded_count += 1
                            sys.stdout.write('\rUploaded manifest file {} of {}'.format(uploaded_count, files_to_upload))
                            sys.stdout.flush()
                            break
                        except IOError:
                            no_manifest.add(validation_manifest.local_file_name)
                        except FileNotFoundError:
                            no_manifest.add(validation_manifest.local_file_name)

                        except json.decoder.JSONDecodeError as e:
                            error = 'JSON Error: There was an error in your json file: {}\nPlease review and try again: {}\n'.format \
                                (validation_manifest.local_file_name, e)
                            raise Exception(error)

        for file in no_manifest:
            logger.info('Manifest file not found: %s',file)

        if no_manifest:
            logger.info(
                '\nYou must make sure all manifest files listed in your validation file'
                ' are located in the specified directory. Please try again.')

            retry = input('Press the "Enter" key to specify location(s) for manifest files and try again:')
            self.manifest_path = retry.split(' ')
            self.process_manifests(r, yes_manifest=yes_manifest, validation_results=self.validation_result)
        else:
            logger.info('\r\nFinished uploading all manifest files')
        while not self.validation_result.status.startswith(Status.COMPLETE):
            time.sleep(1.1)
            response = get_request("/".join([self.api, r['id']]), auth=self.auth)
            self.validation_result = Validation.ValidationManifestResult(response, self)
            for m in self.validation_result.manifests:
                if m.status == Status.ERROR:
                    error = 'JSON Error: There was an error in your json file: {}\nPlease review and try again: {}\n'.format(
                        m.local_file_name, m.errors[0])
                    raise Exception(error)

    class ValidationManifest:


        def __init__(self, _manifest, validation):
            self.config = validation.config
            self.status = _manifest['status']
            self.local_file_name = _manifest['localFileName']
            self.manifestUuid = _manifest['manifestUuid']
            self.errors = _manifest['errors']
            self.url = _manifest['_links']['self']['href']
            self.hide_progress = validation.hide_progress
            self.auth = validation.auth

        def upload_manifest(self, _fp):
            manifest_object = json.load(_fp)
            put_request(self.url, payload=manifest_object, auth=self.auth)

    class ValidationManifestResult:

        def __init__(self, _validation_result, validation):
            self.done = _validation_result['done']
            self.id = _validation_result['id']
            self.status = _validation_result['status']
            self.errors = _validation_result['errors']
            self.short_name = _validation_result['short_name']
            self.scope = _validation_result['scope']
            manifests = []
            for _manifest in _validation_result['manifests']:
                manifests.append(Validation.ValidationManifest(_manifest, validation))
            self.manifests = manifests

    class ValidationTask(threading.Thread, Protocol):
        def __init__(self, file_queue, result_queue, api, scope, responses, validation_progress, exit, validation_timeout, auth):
            threading.Thread.__init__(self)
            self.file_queue = file_queue
            self.result_queue = result_queue
            self.api = api
            self.api_scope = scope
            self.responses = responses
            self.progress_bar = validation_progress
            self.shutdown_flag = threading.Event()
            self.exit = exit
            self.validation_timeout = validation_timeout
            self.auth = auth

        @staticmethod
        def get_protocol(cls):
            return cls.CSV

        def run(self):
            while True and not self.shutdown_flag.is_set():
                polling = 0
                file_name = self.file_queue.get()
                if file_name == "STOP":
                    self.file_queue.put("STOP")
                    self.shutdown_flag.set()
                    break
                try:
                    file = open(file_name, 'r')
                except IOError:
                    if self.progress_bar:
                        self.progress_bar.close()
                    message = 'This file does not exist in current directory: {}'.format(file_name)
                    logger.error(message)

                    exit_client()

                data = file.read()

                response = post_request(self.api_scope, data, timeout=self.validation_timeout, headers = {'content-type':'text/csv'}, auth=self.auth)
                while response and not response['done']:
                    response = get_request("/".join([self.api, response['id']]), auth=self.auth)
                    time.sleep(polling)
                    polling += 1
                if response:
                    response = get_request("/".join([self.api, response['id']]), auth=self.auth)
                    self.result_queue.put((response, file_name))
                    if self.progress_bar:
                        self.progress_bar.update(n=1)
                # Stop thread after adding validation response
                self.file_queue.task_done()


class Status:
    UPLOADING = 'Uploading'
    SYSERROR = 'SystemError'
    PENDING_MANIFEST = 'PendingManifestFiles'
    COMPLETE = 'Complete'
    ERROR = 'Error'



import logging

from NDATools.clientscripts.vtcmd import configure, parse_args, main, validate_files
from NDATools import *
from unittest import TestCase
try:
    import mock
except ImportError:
    from mock import patch


class TestValidation(TestCase):
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s:%(levelname)s:%(message)s")

    def test_manifest_validation_only_success(self):
        test_file = r'test_files/image03_manifest/image03_w_populated_manifest_empty_image_file.csv'
        test_directory = r'test_files/image03_manifest'
        test_manifest_directory = r'test_files/image03_manifest'
        worker_threads = 1

        testargs = ['files', test_file, '-l', test_directory, '-m', test_manifest_directory, '-wt', str(worker_threads)]

        with mock.patch.object(sys, 'argv', testargs):
            args = parse_args()
            config = configure(args)
            self.assertEqual(
                args.files[0].encode('unicode-escape').decode(),
                test_file.encode('unicode-escape').decode())
            self.assertTrue(config.manifest_path)
            self.assertFalse(args.buildPackage)
            self.assertFalse(config.skip_local_file_check)
            main()

    def test_manifest_validation_build_package_config(self):
        test_file = r'test_files/image03_manifest/image03_w_populated_manifest_empty_image_file.csv'
        test_directory = r'test_files/image03_manifest'
        test_manifest_directory = r'test_files/image03_manifest'
        test_title = 'nda-tools-validation-integration-test'
        test_description = 'nda-tools-validation-integration-test'
        test_collection_id = 17
        worker_threads = 1

        testargs = ['files', test_file, '-l', test_directory, '-m', test_manifest_directory, '-b', '-t',
                    test_title, '-d', test_description, '-c', str(test_collection_id), '--skipLocalAssocFileCheck',
                    '-wt', str(worker_threads)]

        with mock.patch.object(sys, 'argv', testargs):
            args = parse_args()
            config = configure(args)
            w = False
            bp = False
            validation_results = validate_files(args.files, w, bp, threads=args.workerThreads, config=config)
            self.assertIsNotNone(validation_results[0])
            self.assertIsNotNone(validation_results[1])
            logging.debug(validation_results)
            return validation_results, config

    def test_s3_validation(self):
        test_file = r'test_files/MultipleDataTypes/image03_mixed_file_size_test_case_stage.csv'
        test_s3_bucket = 'NDAR_Central_1_dev'
        test_s3_prefix = 'submission_23596'
        test_title = 'nda-tools-validation-integration-test'
        test_description = 'nda-tools-validation-integration-test'
        test_collection_id = 17
        worker_threads = 1

        testargs = ['files', test_file, '-s3', test_s3_bucket, '-pre', test_s3_prefix, '-b', '-t', test_title,
                    '-d', test_description,'-c', str(test_collection_id), '-wt', str(worker_threads)]

        with mock.patch.object(sys, 'argv', testargs):
            args = parse_args()
            config = configure(args)
            w = False
            bp = False
            validation_results = validate_files(args.files, w, bp, threads=args.workerThreads, config=config)
            self.assertIsNotNone(validation_results[0])
            self.assertIsNotNone(validation_results[1])
            print(validation_results)
            logging.debug(validation_results)
            return validation_results, config



from __future__ import absolute_import, with_statement

import copy
import datetime
from shutil import copyfile
import sys
import uuid

from boto3.s3.transfer import TransferConfig
from requests import HTTPError

import NDATools
from NDATools import Utils
from NDATools.AltEndpointSSLAdapter import AltEndpointSSLAdapter

IS_PY2 = sys.version_info < (3, 0)

if IS_PY2:
    from Queue import Queue
    from io import open
    from urllib import quote_plus
else:
    from queue import Queue
    from urllib.parse import quote_plus
import csv
from threading import Thread
import multiprocessing
from NDATools.Utils import *
import requests

logger = logging.getLogger(__name__)


class Worker(Thread):
    """ Thread executing tasks from a given tasks queue """

    def __init__(self, tasks):
        Thread.__init__(self)
        self.tasks = tasks
        self.daemon = True
        self.start()

    def run(self):
        while True:
            func, args, kargs = self.tasks.get()
            try:
                func(*args, **kargs)
            except Exception as e:
                # An exception happened in this thread
                logger.info(str(e))
                logger.info(get_traceback())

            finally:
                # Mark this task as done, whether an exception happened or not
                self.tasks.task_done()


class ThreadPool:
    """ Pool of threads consuming tasks from a queue """

    def __init__(self, num_threads, queue_size=None):
        queue_size = queue_size or num_threads * 100
        self.tasks = Queue(queue_size)
        for _ in range(num_threads):
            Worker(self.tasks)

    def add_task(self, func, *args, **kargs):
        """ Add a task to the queue """
        self.tasks.put((func, args, kargs))

    def map(self, func, args_list):
        """ Add a list of tasks to the queue """
        for args in args_list:
            self.add_task(func, args)

    def wait_completion(self):
        """ Wait for completion of all the tasks in the queue """
        self.tasks.join()


class Download(Protocol):

    def __init__(self, download_config, args):

        # Instance variables from config
        self.config = download_config
        self.package_url = self.config.package_api
        self.datadictionary_url = self.config.datadictionary_api
        self.username = download_config.username
        self.password = download_config.password
        self.auth = requests.auth.HTTPBasicAuth(self.config.username, self.config.password)

        # Instance Variables from 'args'
        if args.directory:
            download_directory = args.directory[0]
        else:
            download_directory = os.path.join(NDATools.NDA_TOOLS_DOWNLOADS_FOLDER, str(args.package))
        self.downloadcmd_package_metadata_directory = os.path.join(NDATools.NDA_TOOLS_DOWNLOADS_FOLDER,
                                                                   str(args.package))
        self.package_download_directory = Utils.convert_to_abs_path(download_directory)
        self.s3_links_file = args.txt
        self.inline_s3_links = args.paths
        self.package_id = args.package
        self.data_structure = args.datastructure
        self.thread_num = args.workerThreads if args.workerThreads else max([1, multiprocessing.cpu_count() - 1])
        self.regex_file_filter = args.file_regex
        if self.s3_links_file:
            self.download_mode = 'text'
        elif self.data_structure:
            self.download_mode = 'datastructure'
        elif self.inline_s3_links:
            self.download_mode = 'paths'
        else:
            self.download_mode = 'package'
        self.verify_flg = args.verify

        if not self.verify_flg and not args.workerThreads:
            logger.warning('\nNo value specified for --workerThreads. Using the default option of {}'.format(
                self.thread_num))
            logger.info(
                'Important - You can configure the thread count setting using the --workerThreads argument to maximize your download speed.\n')
        # for copying files directly to another s3 bucket
        self.custom_user_s3_endpoint = args.s3_destination

        # non-configurable default instance variables
        self.download_queue = Queue()
        self.local_file_names = {}  # map of package-file-id to alias
        self.package_file_download_errors = set()
        # self.package_file_download_errors needs a lock if multiple threads will be adding to it simultaneously
        self.package_file_download_errors_lock = threading.Lock()

        self.download_job_uuid = None

        self.download_job_manifest_column_defs = {
            'uuid': self.download_job_uuid,
            'run_date': time.strftime("%Y%m%dT%H%M%S"),
            'package_id': self.package_id,
            'download_directory': self.package_download_directory,
            's3_destination': self.custom_user_s3_endpoint,
            'data_structure': self.data_structure,
            's3_links_file': self.s3_links_file,
            'regex': self.regex_file_filter
        }
        self.download_job_progress_report_column_defs = {
            'package_file_id': None,
            'package_file_expected_location': None,
            'nda_s3_url': None,
            'exists': False,
            'expected_file_size': None,
            'actual_file_size': 0,
            'e_tag': None,
            'download_complete_time': None
        }
        self.download_progress_report_file_path = self.initialize_verification_files()
        self.default_download_batch_size = 50

    # exlcude arg list is the long-parameter name
    def build_rerun_download_cmd(self, exclude_arg_list):
        download_cmd = 'downloadcmd -dp {}'.format(self.package_id)

        if self.download_mode == 'text' and '--txt' not in exclude_arg_list:
            download_cmd += ' -t {}'.format(self.s3_links_file)
        if self.download_mode == 'datastructure' and '--datastructure' not in exclude_arg_list:
            download_cmd += ' -ds {}'.format(self.data_structure)
        if self.regex_file_filter and '--file-regex' not in exclude_arg_list:
            download_cmd += ' --file-regex {}'.format(self.regex_file_filter)
        if self.username and '--username' not in exclude_arg_list:
            download_cmd += ' -u {}'.format(self.username)
        if self.package_download_directory and '--directory' not in exclude_arg_list:
            download_cmd += ' -d {}'.format(self.package_download_directory)
        if self.verify_flg and '--verify' not in exclude_arg_list:
            download_cmd += ' --verify'
        if self.thread_num and '--workerThreads' not in exclude_arg_list:
            download_cmd += ' -wt {}'.format(self.thread_num)
        if self.custom_user_s3_endpoint and '--s3-destination' not in exclude_arg_list:
            download_cmd += ' -s3 {}'.format(self.custom_user_s3_endpoint)

        return download_cmd

    def start(self):
        logger.info('')
        logger.info('Getting Package Information...')
        package_resource = self.get_package_info()
        logger.info('')
        logger.info('Package-id: {}'.format(self.package_id))
        logger.info('Name: {}'.format(package_resource['description']))
        logger.info('Has associated files?: {}'.format('Yes' if package_resource['has_associated_files'] else 'No'))
        # Dont print this out because at the moment, the number coming back from the service includes duplicates.
        # uncomment when that is fixed
        print('Number of files in package: {}'.format(package_resource['file_count']))
        logger.info('Total Package Size: {}'.format(Utils.human_size(package_resource['total_package_size'])))
        logger.info('')

        files = []
        if self.download_mode == 'datastructure':
            logger.info('Downloading S3 links from data structure: {}'.format(self.data_structure))
            if not package_resource['has_associated_files']:
                logger.info(''''No Associated files detected in this package. In order to download associated files, you must create a new package
        on the NDA website and make sure that you check the option to "Include associated files"''')
                exit_client()
            files = self.use_data_structure()
        elif self.download_mode == 'text':
            logger.info('Downloading S3 links from text file: {}'.format(self.s3_links_file))
            files = self.use_s3_links_file()
        elif self.download_mode == 'package':
            if self.regex_file_filter:
                logger.info('Downloading files from package {} matching regex {}'.format(self.package_id, self.regex_file_filter))
            else:
                logger.info('Downloading all files from package with id: {}'.format(self.package_id))
        else:
            files = self.query_files_by_s3_path(self.inline_s3_links)

        if files:
            self.local_file_names = {int(f['package_file_id']): f for f in files}
        logger.info('')

        success_files = set()
        download_request_count = 0
        download_start_date = datetime.datetime.now()

        download_progress_report = open(self.download_progress_report_file_path, 'a', newline='')
        download_progress_report_writer = csv.DictWriter(download_progress_report,
                                                         fieldnames=self.download_job_progress_report_column_defs)

        failed_s3_links_file = open(os.path.join(NDATools.NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER,
                                                 'failed_s3_links_file_{}.txt'.format(time.strftime("%Y%m%dT%H%M%S"))),
                                    'a')

        message = 'S3 links for files that failed to download will be written out to {}. You can attempt to download these files later by running: ' \
            .format(failed_s3_links_file.name)
        message += '\n\t{} -t "{}"' \
            .format(self.build_rerun_download_cmd(['--text', '--datastructure']), failed_s3_links_file.name)
        logger.info(message)
        logger.info('')
        time.sleep(1.5)

        if self.download_mode == 'package':
            file_ct = package_resource['file_count']
            file_sz = int(package_resource['total_package_size'])
        else:
            file_ct = len(self.local_file_names.keys())
            file_sz = sum(map(lambda x: x['file_size'], self.local_file_names.values()))

        #remove files that have already been completed
        completed_files = self.get_completed_files_in_download()
        tmp = {int(f['package_file_id']):int(f['actual_file_size']) for f in completed_files}
        completed_file_sz = sum(tmp.values())
        completed_file_ids = set(tmp.keys())
        completed_file_ct = len(completed_file_ids)
        completed_files = tmp = None #remove large structures from memory

        if self.download_mode == 'package' and self.regex_file_filter:
            # cant display file number because its not known
            message = 'Beginning download of files from package matching {} using {} threads'.format(
                self.regex_file_filter,
                self.thread_num)
        else:
            skipping_message = ''
            if completed_file_ct>0:
                download_progress_report_path = os.path.join(self.downloadcmd_package_metadata_directory,
                                                             '.download-progress', self.download_job_uuid,
                                                             'download-progress-report.csv')

                skipping_message = 'Skipping {} files ({}) which have already been downloaded according to log file {}.\n'.format(completed_file_ct, Utils.human_size(completed_file_sz), download_progress_report_path)
                file_ct -= completed_file_ct
                file_sz -= completed_file_sz
            message = '{}Beginning download of {}{} files ({}) to {} using {} threads'.format(
                skipping_message,
                'the remaining ' if skipping_message else '',
                file_ct,
                Utils.human_size(file_sz),
                self.custom_user_s3_endpoint or self.package_download_directory,
                self.thread_num)

        logger.info('')
        logger.info(message)
        time.sleep(5)

        # These are all arrays just so that the print_download_progress_report method can update the variables inside them
        trailing_50_file_bytes = []
        trailing_50_timestamp = [datetime.datetime.now()]


        download_progress_flush_date=[datetime.datetime.now()]
        def write_to_download_progress_report_file(download_record):
            # if file-size =0, there could have been an error. Dont add to file
            if download_record['actual_file_size'] > 0:
                download_progress_report_writer.writerow(download_record)
                if (datetime.datetime.now()-download_progress_flush_date[0]).seconds>10:
                    download_progress_report.flush()
                    download_progress_flush_date[0]=datetime.datetime.now()
                else:
                    pass

        def print_download_progress_report(num_downloaded):

            byte_total = sum(trailing_50_file_bytes)
            download_progress_message = 'Download Progress Report [{}]: \n    {}/{} queued files downloaded so far. ' \
                .format(datetime.datetime.now().strftime('%b %d %Y %H:%M:%S'), num_downloaded, download_request_count)
            download_progress_message += '\n    Last 50 files contained ~ {} bytes and finished in {} (Hours:Minutes:Seconds). ' \
                .format(Utils.human_size(byte_total),
                        str(datetime.datetime.now() - trailing_50_timestamp[0]).split('.')[0])

            seconds_last_50_files = (datetime.datetime.now() - trailing_50_timestamp[0]).seconds
            if seconds_last_50_files == 0:
                seconds_last_50_files = 1  # avoid a 'division by 0' error

            # convert download speed to bits per second
            avg_speed_bps = Utils.human_size((8 * byte_total) // seconds_last_50_files)
            if avg_speed_bps[-1:] == 'B':
                avg_speed_bps = avg_speed_bps.replace('B', 'bps')
            else:
                avg_speed_bps = avg_speed_bps.replace('bytes', 'bps')

            download_progress_message += '\n    Avg download rate (in bits per second) for the last 50 files is ~ {}.' \
                .format(avg_speed_bps)

            download_progress_message += '\n    Download has been in progress for {} (Hours:Minutes:Seconds).\n' \
                .format(str(datetime.datetime.now() - download_start_date).split('.')[0])

            download_progress_message = '\n' + download_progress_message + '\n'
            logger.info(download_progress_message)
            trailing_50_file_bytes.clear()
            trailing_50_timestamp[0] = datetime.datetime.now()

        def download(file_id_to_cred_list):
            package_file_id = file_id_to_cred_list[0]
            cred = file_id_to_cred_list[1]
            # check if  these exist, and if not, get and set:
            download_record = self.download_from_s3link(package_file_id, cred,
                                                        failed_s3_links_file=failed_s3_links_file)
            # dont add bytes if file-existed and didnt need to be downloaded
            if download_record['download_complete_time']:
                trailing_50_file_bytes.append(download_record['actual_file_size'])
            success_files.add(package_file_id)
            num_downloaded = len(success_files)

            if num_downloaded % 50 == 0:
                print_download_progress_report(num_downloaded)

            download_progress_file_writer_pool.add_task(write_to_download_progress_report_file, download_record)

        download_pool = ThreadPool(self.thread_num)
        download_progress_file_writer_pool = ThreadPool(1, 1000)

        for package_file_id_list in self.generate_download_batch_file_ids(completed_file_ids):
            if len(package_file_id_list) > 0:
                additional_file_ct = len(package_file_id_list)
                download_request_count += additional_file_ct
                logger.info('Adding {} files to download queue. Queue contains {} files\n'.format(additional_file_ct,
                                                                                                       download_request_count))
                file_id_to_cred_list = self.get_presigned_urls(list(package_file_id_list))
                download_pool.map(download, file_id_to_cred_list.items())

        download_pool.wait_completion()
        failed_s3_links_file.close()
        download_progress_report.close()

        # dont generate a file if there were no failures
        if not self.package_file_download_errors:
            logger.info('No failures detected. Removing file {}'.format(failed_s3_links_file.name))
            os.remove(failed_s3_links_file.name)

        logger.info('')

        logger.info('Finished processing all download requests @ {}.'.format(datetime.datetime.now()))
        logger.info('     Total download requests {}'
                    .format(download_request_count))
        logger.info('     Total errors encountered: {}'.format(len(self.package_file_download_errors)))

        logger.info('')
        logger.info(' Exiting Program...')

    def use_s3_links_file(self):
        """
        Reads a text file line by line, collects all S3 URLs, and requests the package file ids
        from the package service for those files.
        """
        path_list = set()
        try:
            with open(self.s3_links_file, 'r', encoding='utf-8') as s3_links_file:
                for line in s3_links_file:
                    if line.startswith('s3://'):
                        if self.regex_file_filter and not re.search(self.regex_file_filter, line):
                            pass
                        else:
                            path_list.add(line.strip())
            if not path_list:
                logger.info(
                    'No valid paths found in s3-links file. If you specified a regular expression, make sure your regular expression is correct before re-running the command.')
                exit_client()

            return self.query_files_by_s3_path(path_list)
        except IOError as e:
            logger.error(
                '{} not found. Please enter the correct path to your file and try again.'.format(self.data_structure))
            raise e

    def use_data_structure(self):
        """
        Downloads a data structure file from the package, the data_structure_manifest file,
        parses both files to identify all associated files for the data structure, and then
        requests the package file ids for those files from the package service
        """
        # TODO - add paging in case the number of files is large
        path_list = self.get_files_from_datastructure(self.data_structure)
        if not path_list:
            logger.info(
                '{} data structure is not included in the package {}'.format(self.data_structure, self.package_id))
            exit_client()
        if self.regex_file_filter:
            path_list = list(filter(lambda x: re.search(self.regex_file_filter, x['download_alias']), path_list))

        if not path_list:
            logger.info(
                'No valid paths found in data-structure file. If you specified a regular expression, make sure your regular expression is correct before re-running the command.')
            exit_client()
        return path_list

    def download_from_s3link(self, package_file_id, s3_link, err_if_exists=False, failed_s3_links_file=None):

        # check if we are downloading from alt endpoint where bucket name contains dots.
        def get_http_adapter(s3_link):
            bucket, path = deconstruct_s3_url(s3_link)
            config = {'max_retries': 10}
            if ('.' in bucket):
                return AltEndpointSSLAdapter(**config)
            return HTTPAdapter(**config)

        # use this instead of exists_ok in order to work with python v.2
        def mk_dir_ignore_err(dir):
            try:
                os.makedirs(os.path.normpath(dir))
            except FileExistsError as e:
                pass
            except OSError as e:
                # Raise exception for any errors other than FileExists error
                # Using OSError for version compatibility
                if e.errno != 17:
                    raise
                pass

        # declare to avoid 'reference before declared' error
        source_uri = None
        bytes_written = 0
        return_value = copy.deepcopy(self.download_job_progress_report_column_defs)
        return_value['package_file_id'] = str(package_file_id)
        try:
            alias = self.local_file_names[package_file_id]['download_alias']
            return_value['expected_file_size'] = self.local_file_names[package_file_id]['file_size']
            return_value['package_file_expected_location'] = alias
            completed_download = os.path.normpath(os.path.join(self.package_download_directory, alias))
            partial_download = os.path.normpath(
                os.path.join(self.package_download_directory, alias + '.partial'))
            downloaded = False
            resume_header = None

            if not self.custom_user_s3_endpoint:
                if os.path.isfile(completed_download):
                    if err_if_exists:
                        msg = "File {} already exists. Move or rename the file before re-running the command to continue".format(
                            completed_download)
                        logger.info(msg)
                        logger.info('Exiting...')
                        sys.exit(1)

                    logger.info('Skipping download (already exists): {}'.format(completed_download))
                    actual_size = os.path.getsize(completed_download)
                    return_value['actual_file_size'] = actual_size
                    return_value['exists'] = True
                    return return_value

                if os.path.isfile(partial_download):
                    downloaded = True
                    downloaded_size = os.path.getsize(partial_download)
                    resume_header = {'Range': 'bytes={}-'.format(downloaded_size)}
                    logger.info('Resuming download: {}'.
                                format(partial_download))
                else:
                    mk_dir_ignore_err(os.path.dirname(partial_download))
                    logger.info('Starting download: {}'.format(partial_download))

            if self.custom_user_s3_endpoint:
                # downloading directly to s3 bucket
                # get cred for file
                response = self.get_temp_creds_for_file(package_file_id, self.custom_user_s3_endpoint)
                ak = response['access_key']
                sk = response['secret_key']
                sess_token = response['session_token']
                source_uri = response['source_uri']
                dest_uri = response['destination_uri']

                dest_bucket, dest_path = Utils.deconstruct_s3_url(dest_uri)
                src_bucket, src_path = Utils.deconstruct_s3_url(source_uri)

                logger.info('Starting download: s3://{}/{}'.format(dest_bucket, dest_path))

                # boto3 copy
                sess = boto3.session.Session(aws_access_key_id=ak,
                                             aws_secret_access_key=sk,
                                             aws_session_token=sess_token,
                                             region_name='us-east-1')

                s3_client = sess.client('s3')
                response = s3_client.head_object(Bucket=src_bucket, Key=src_path)
                return_value['actual_file_size'] = response['ContentLength']
                return_value['e_tag'] = response['ETag'].replace('"', '')
                return_value['nda_s3_url'] = 's3://{}/{}'.format(src_bucket, src_path)

                s3 = sess.resource('s3')
                copy_source = {
                    'Bucket': src_bucket,
                    'Key': src_path
                }

                def print_upload_part_info(bytes):
                    logger.info('Transferred {} for {}'.format(Utils.human_size(bytes), return_value['nda_s3_url']))

                KB = 1024
                MB = KB * KB
                GB = KB ** 3
                LARGE_OBJECT_THRESHOLD = 5 * GB
                args = {
                    'ExtraArgs': {'ACL': 'bucket-owner-full-control'}
                }

                if int(return_value['actual_file_size']) >= LARGE_OBJECT_THRESHOLD:
                    logger.info('Transferring large object {} ({}) in multiple parts'
                                .format(return_value['nda_s3_url'],
                                        Utils.human_size(int(return_value['actual_file_size']))))
                    config = TransferConfig(multipart_threshold=LARGE_OBJECT_THRESHOLD, multipart_chunksize=1 * GB)
                    args['Config'] = config
                    args['Callback'] = print_upload_part_info

                s3.meta.client.copy(copy_source,
                                    dest_bucket,
                                    dest_path,
                                    **args)

            else:
                # downloading to local machine

                with requests.session() as s:
                    s.mount(s3_link, get_http_adapter(s3_link))
                    if resume_header:
                        s.headers.update(resume_header)
                    with open(partial_download, "ab" if downloaded else "wb") as download_file:
                        with s.get(s3_link, stream=True) as response:
                            response.raise_for_status()
                            for chunk in response.iter_content(chunk_size=1024 * 1024 * 5):  # iterate 5MB chunks
                                if chunk:
                                    bytes_written += download_file.write(chunk)
                os.rename(partial_download, completed_download)
                logger.info('Completed download {}'.format(completed_download))
                return_value['actual_file_size'] = bytes_written
                bucket, key = Utils.deconstruct_s3_url(s3_link)
                return_value['nda_s3_url'] = 's3://{}/{}'.format(bucket, key)
            return_value['exists'] = True
            return_value['download_complete_time'] = time.strftime("%Y%m%dT%H%M%S")
            return return_value

        except Exception as e:
            if not s3_link and not source_uri:
                # we couldnt get credentials, which means the service has become un-responsive.
                # Instruct the user to retry at another time
                logger.info('')
                logger.info(
                    'Unexpected Error During File Download - Service Unresponsive. Unable to obtain credentials for file-id {}'.format(
                        package_file_id))
                logger.info('Please re-try downloading files at a later time. ')
                logger.info('You may contact NDAHelp@mail.nih.gov for assistance in resolving this error.')
                # use os._exit to kill the whole program. This works even if this is called in a child thread, unlike sys.exit()
                os._exit(1)

            self.write_to_failed_download_link_file(failed_s3_links_file, s3_link=s3_link, source_uri=source_uri)

            error_code = -1 if not isinstance(e, HTTPError) else int(e.response.status_code)
            if error_code == 404:
                message = 'This path is incorrect: {}. Please try again.'.format(s3_link)
                logger.error(message)
            elif error_code == 403:
                message = '\nThis is a private bucket. Please contact NDAR for help: {}'.format(s3_link)
                logger.error(message)
            else:
                logger.error(str(e))
                logger.error(get_traceback())
                if 'operation: Access Denied' in str(e):
                    logger.error('')
                    logger.error(
                        'This error is likely caused by a misconfiguration on the target s3 bucket')
                    logger.error(
                        "For more information about how to correctly configure the target bucket, run 'downloadcmd -h' and read the description of the s3 argument")
                    logger.error('')
                    time.sleep(2)

            # if source_uri is set, it means they're downloading to s3 bucket and there will not be any partial file
            if bytes_written == 0 and not source_uri:
                try:
                    os.remove(partial_download)
                except:
                    logger.error('error removing partial file {}'.format(partial_download))
            return return_value

    def write_to_failed_download_link_file(self, failed_s3_links_file, s3_link, source_uri):
        src_bucket, src_path = Utils.deconstruct_s3_url(s3_link if s3_link else source_uri)
        s3_address = 's3://' + src_bucket + '/' + src_path

        with self.package_file_download_errors_lock:
            self.package_file_download_errors.add(s3_address)
            if failed_s3_links_file:
                failed_s3_links_file.write(s3_address + "\n")
                failed_s3_links_file.flush()

    def generate_download_batch_file_ids(self, completed_file_ids):
        if self.download_mode == 'package':
            already_downloaded = len(completed_file_ids)

            #  write generator function that goes through each page in file listing
            #  and yields file-ids for files in package. Before returning, self.local_file_path must be set
            page = (already_downloaded // self.default_download_batch_size) + 1
            while True:
                logger.debug(f'getting package files for package {self.package_id} and page {page} at time {datetime.datetime.now()}')
                files = self.get_package_files_by_page(page, self.default_download_batch_size)
                logger.debug(f'finished getting package files for package {self.package_id} and page {page} at time {datetime.datetime.now()}')
                tmp = {r['package_file_id']: r for r in files if int(r['package_file_id']) not in completed_file_ids}
                self.local_file_names.update(tmp)
                package_file_ids = tmp.keys()
                page += 1
                if not package_file_ids:
                    break
                yield package_file_ids

        else:

            package_file_list = [f for f in {k:v for k,v in self.local_file_names.items() if int(k) not in completed_file_ids}]
            batches = [package_file_list[i:i + self.default_download_batch_size] for i in range(0, len(package_file_list), self.default_download_batch_size)]
            for batch in batches:
                yield batch

    def find_matching_download_job(self, download_job_manifest_path):
        def is_job_match(possible_match):
            must_match = [
                'data_structure',
                'download_directory',
                's3_destination',
                's3_links_file',
                'package_id',
                'regex'
            ]

            def test_match(key):
                # None gets converted to empty string.
                # Convert it back if empty string is detected so that None values can be compared using == operator
                val1 = possible_match[key] or None
                # values from download_job_manifest_column_defs will never be None, instead they will be an empty string ''
                val2 = self.download_job_manifest_column_defs[key]
                if key == 'download_directory':
                    val2 = Utils.convert_to_abs_path(val2)
                    val1 = Utils.convert_to_abs_path(val1)
                elif key == 's3_links_file':
                    # only convert to basename if values are specified for both
                    if val2 and val1:
                        val2 = os.path.basename(val2)
                        val1 = os.path.basename(val1)
                return val1 == val2

            return all(map(test_match, must_match))

        # FIND MATCHING JOB RECORD IF IT EXISTS
        with open(download_job_manifest_path, newline='') as csvfile:
            job_reader = csv.DictReader(csvfile)
            for job in job_reader:
                if is_job_match(job):
                    return job
        return None

    def initialize_verification_files(self):

        download_job_manifest_columns = self.download_job_manifest_column_defs.keys()

        def initialize_download_progress_file(fp):
            with open(fp, 'a', newline='') as file:
                writer = csv.DictWriter(file, fieldnames=self.download_job_progress_report_column_defs)
                writer.writeheader()

        def add_entry_to_job_manifest(fp):
            self.download_job_uuid = self.download_job_manifest_column_defs['uuid'] = str(uuid.uuid4())
            with open(fp, 'a', newline='') as file:
                writer = csv.DictWriter(file, fieldnames=download_job_manifest_columns)
                writer.writerow(self.download_job_manifest_column_defs)

        def initialize_job_manifest_file(fp):
            with open(fp, 'w', newline='') as file:
                writer = csv.DictWriter(file, fieldnames=download_job_manifest_columns)
                writer.writeheader()

        if not os.path.exists(self.downloadcmd_package_metadata_directory):
            os.mkdir(self.downloadcmd_package_metadata_directory)

        DOWNLOAD_PROGRESS_FOLDER = os.path.join(self.downloadcmd_package_metadata_directory, '.download-progress')
        if not os.path.exists(DOWNLOAD_PROGRESS_FOLDER):
            os.mkdir(DOWNLOAD_PROGRESS_FOLDER)

        download_job_manifest_path = os.path.join(DOWNLOAD_PROGRESS_FOLDER, 'download-job-manifest.csv')
        if not os.path.exists(download_job_manifest_path):
            initialize_job_manifest_file(download_job_manifest_path)

        job_record = self.find_matching_download_job(download_job_manifest_path)
        if job_record is not None:
            self.download_job_uuid = job_record['uuid']

        if not self.download_job_uuid:
            add_entry_to_job_manifest(download_job_manifest_path)

        DOWNLOAD_JOB_UUID_DIR = os.path.join(DOWNLOAD_PROGRESS_FOLDER, str(self.download_job_uuid))
        if not os.path.exists(DOWNLOAD_JOB_UUID_DIR):
            os.mkdir(DOWNLOAD_JOB_UUID_DIR)

        download_progress_report_file = os.path.join(DOWNLOAD_JOB_UUID_DIR, 'download-progress-report.csv')
        if not os.path.exists(download_progress_report_file):
            initialize_download_progress_file(download_progress_report_file)
        return download_progress_report_file

    '''        
            Steps -
            1. Find path of correct download-progress-report file from .download-progress/download-job-manifest.csv (if it exists),
                for the given target directory and download mode
            2. Make a copy of the download-progress-report and save into verification folder. Name it 'download-verification-report.csv'
            3. Consider everything currently in the download-verification-report.csv where expected file-size=actual file-size as being downloaded
                     a. there really shouldnt be any entries where the expected filesize doesnt match actual size , but run the code anyway
            4. Read file-names of download-verification-report.csv into set() in memory
            *5. Get the complete file-listing for the download (using the provided arguments -d, -ds, -t and -dp)
            6. Add anything that is not in the set in step 4 into the download-verification-report.csv
            7. Run os.path.exists() on each added entry - if it exists, add file-size info to download-verification-report.csv. if not, set
            exists columns and file-size columns
            8. Run through the file download-verification-report.csv and collect everything where the actual and expected file-sizes dont match up.
            9. Create the download-verification-retry-s3-links.csv file from the entries found in step 8
    '''

    def verify_download(self):

        if self.custom_user_s3_endpoint:
            raise Exception(
                'The --verify command does not yet support checking for files in s3 endpoints. This feature will be added in a future iteration...')
            exit_client()

        verification_report_path = os.path.join(self.downloadcmd_package_metadata_directory,
                                                'download-verification-report.csv')
        err_mess_template = 'Cannot start verification process - {} already exists \nYou must move or rename the file in order to continue'
        if os.path.exists(verification_report_path):
            logger.info('')
            logger.info(err_mess_template.format(verification_report_path))
            exit_client()

        fpath = os.path.join(self.downloadcmd_package_metadata_directory, 'download-verification-retry-s3-links.csv')
        if os.path.exists(fpath):
            logger.info(err_mess_template.format(fpath))
            exit_client()

        def get_download_progress_report_path():
            return os.path.join(self.downloadcmd_package_metadata_directory, '.download-progress',
                                self.download_job_uuid, 'download-progress-report.csv')

        def parse_download_progress_report_for_files(download_progress_report_path):
            files = []
            if os.path.exists(download_progress_report_path):
                with open(download_progress_report_path, newline='') as csvfile:
                    file_reader = csv.DictReader(csvfile)
                    files = [f for f in file_reader]
            return files

        def get_complete_file_list():
            if self.download_mode in ['text', 'datastructure']:
                if self.download_mode == 'datastructure':
                    files = self.use_data_structure()
                elif self.download_mode == 'text':
                    files = self.use_s3_links_file()
                self.local_file_names = {int(f['package_file_id']): f for f in files}
                return set(self.local_file_names)
            elif self.download_mode == 'package':
                logger.info('Getting list of all files in package. If your package is large, this may take some time')
                page = 1
                batch_size = 1000
                all_results = []
                while True:
                    results = self.get_package_files_by_page(page, batch_size)
                    aliases = {r['package_file_id']: r for r in results}
                    self.local_file_names.update(aliases)
                    if not results:
                        break
                    else:
                        all_results.append(results)
                        logger.info('Retrieved {} 1000 files. At file #{}'.format('first' if page == 1 else 'next',
                                                                                  ((page - 1) * batch_size) + 1))
                    page += 1

                return set(self.local_file_names)
            else:
                raise Exception('Unsupported download mode: {}'.format(self.download_mode))

        def create_download_verification_retry_links_file(s3_links):
            fpath = os.path.join(self.downloadcmd_package_metadata_directory,
                                 'download-verification-retry-s3-links.csv')
            with open(fpath, 'w') as retry_file:
                for link in s3_links:
                    retry_file.write(link + '\n')

        def add_files_to_report(download_progress_report_path, verification_report_path, probably_missing_files_list):
            copyfile(download_progress_report_path, verification_report_path)

            all_records = []
            for file_id in probably_missing_files_list:
                file_info = self.local_file_names[file_id]
                record = copy.deepcopy(self.download_job_progress_report_column_defs)
                # TODO - consider making these the same names
                record['package_file_expected_location'] = file_info['download_alias']
                record['expected_file_size'] = file_info['file_size']
                record['package_file_id'] = file_id
                download_path = os.path.join(self.package_download_directory,
                                             self.local_file_names[file_id]['download_alias'])
                if os.path.exists(download_path):
                    record['exists'] = True
                    stat = os.stat(download_path)
                    record['actual_file_size'] = stat.st_size
                    if record['actual_file_size'] == record['expected_file_size']:
                        record['download_complete_time'] = time.strftime("%Y%m%dT%H%M%S", time.localtime(stat.st_ctime))

                all_records.append(record)

            # batch requests to reduce overhead
            # TODO - add s3 location to file resource in order to eliminate this step?
            logger.info('Retrieving s3 url information for {} nda file records'.format(len(all_records)))
            batch_size = 1000
            batches = [all_records[i:i + batch_size] for i in range(0, len(all_records), batch_size)]
            for batch in batches:
                result = self.get_presigned_urls([record['package_file_id'] for record in batch])
                # result is a dictionary of package-file-id to presignedUrl
                for record in batch:
                    ps_url = result[record['package_file_id']]
                    dest_bucket, dest_path = Utils.deconstruct_s3_url(ps_url)
                    record['nda_s3_url'] = 's3://{}/{}'.format(dest_bucket, dest_path)

            with open(verification_report_path, 'a', newline='') as verification_report:
                download_progress_report_writer = csv.DictWriter(verification_report,
                                                                 fieldnames=self.download_job_progress_report_column_defs)
                download_progress_report_writer.writerows(all_records)

            return [record['nda_s3_url'] for record in all_records if
                    record['actual_file_size'] < record['expected_file_size']]

        logger.info('')
        logger.info(
            'Running verification process. This process will check whether all of the files from the following downloadcmd were successfully downloaded to the computer:')

        verification_report_path = os.path.join(self.downloadcmd_package_metadata_directory,
                                                'download-verification-report.csv')

        logger.info('{}'.format(self.build_rerun_download_cmd(['--verify'])))
        logger.info('')
        pr_path = get_download_progress_report_path()
        logger.info('Getting expected file list for download...')
        complete_file_set = get_complete_file_list()

        # Sometimes there are dupes in the qft table. eliminate to get accurate file count
        accurate_file_ct = len(set(map(lambda x: x['download_alias'], self.local_file_names.values())))
        file_sz = Utils.human_size(sum(map(lambda x: x['file_size'], self.local_file_names.values())))

        logger.info(
            '{} files are expected to have been downloaded from the command above, totaling {}'.format(accurate_file_ct,
                                                                                                       file_sz))
        logger.info('')
        logger.info('Parsing program system logs for history of completed downloads...')
        logger.info(
            'Important - if you think files may have been deleted from your system after the download was run, you should remove the system log at {}'
            ' and re-run the --verify command. This will force the program to check for these files instead of assuming they exist based on system log entries. This will cause the --verify step to take longer'
            ' to finish but will be necessary for accurate results.'.format(pr_path))
        downloaded_file_records = parse_download_progress_report_for_files(pr_path)
        # There shouldn't be duplicates in the system logs, but check anyway
        downloaded_file_records_count = len({f['package_file_expected_location'] for f in downloaded_file_records})
        downloaded_file_set = {int(f['package_file_id']) for f in downloaded_file_records}
        logger.info('')
        logger.info(
            'Found {} complete file downloads according to log file {}'.format(downloaded_file_records_count, pr_path))
        logger.info('')
        probably_missing_files = complete_file_set - downloaded_file_set
        logger.info(
            'Checking {} for all files which were not found in the program system logs. Detailed report will be created at {}...'
            .format(self.package_download_directory, verification_report_path))
        undownloaded_s3_links = add_files_to_report(pr_path, verification_report_path, probably_missing_files)
        logger.info('')
        if undownloaded_s3_links:
            logger.info(
                'Finished verification process and file check. Found {} files that were missing or whose size on disk were less than expected'.format(
                    len(undownloaded_s3_links)))
            logger.info('')
            logger.info(
                'Generating list of s3 links for all missing/incomplete files...'.format(len(undownloaded_s3_links)))
            create_download_verification_retry_links_file(undownloaded_s3_links)
            incomplete_s3_fp = os.path.join(self.downloadcmd_package_metadata_directory,
                                            'download-verification-retry-s3-links.csv')
            logger.info(
                'Finished creating {} file. \nThis file contains s3-links for all files that were found to be missing or incomplete. You may '
                'download these files by running:\n'
                '   {} -t {}'.format(incomplete_s3_fp,
                                     self.build_rerun_download_cmd(['--verify', '--text', '--datastructure']),
                                     incomplete_s3_fp))
        else:
            logger.info(
                'Finished verification process and file check. No missing files found. All files match expected size. Download 100% complete.')

        logger.info('')
        logger.info(
            'Details about status of files in download can be found at {} (This file can be opened with Excel or Google Spreadsheets)'.format(
                verification_report_path))
        exit_client()

        # Step 5 should be made into a modular function because creating a file listing per download is something the user might be interested in having
        # potential future feature of the tool.

        # Note 1 - when implementing this for s3, we will need to add instructions to provide the list-bucket permission to user
        # Note 2 - eventually, we might be able to add the e-tag data-attribute and start verifying download content by comparing the e-tag attribute

    def query_files_by_s3_path(self, path_list):
        if not path_list:
            exit_client(signal=signal.SIGTERM, message='Illegal Argument - path_list cannot be empty')
        url = self.package_url + '/{}/files'.format(self.package_id)
        try:
            response = post_request(url, list(path_list), auth=self.auth,
                                    error_handler=HttpErrorHandlingStrategy.reraise_status,
                                    deserialize_handler=DeserializeHandler.none)
            response.raise_for_status()
            return response.json()
        except HTTPError as e:
            if e.response.status_code == 404:
                message = e.response.text
                invalid_s3_links = set(map(lambda x: x.rstrip(), message.split('\n')[1:]))
                logger.info('WARNING: The following associated files were not found in the package '
                            'and will not be downloaded\n{}'.format('\n'.join(invalid_s3_links)))
                logger.info('')
                for i in invalid_s3_links:
                    path_list.remove(i)

                if not path_list:
                    logger.info('All files requested for download were invalid.')
                    logger.info('This error may be encountered if you did not select the "include associated files" '
                                'option when before your data package. If you wish to download associated files, you will need to create a new '
                                'package')
                    logger.info(
                        'If you are sure you created your package with this option selected, please contact NDAHelp@mail.nih.gov '
                        'for assistance in resolving this error.')
                    sys.exit(1)
                else:
                    return self.query_files_by_s3_path(path_list)  # retry request after excluding the invalid s3links
            else:
                HttpErrorHandlingStrategy.print_and_exit(e.response)

    def get_temp_creds_for_file(self, package_file_id, custom_user_s3_endpoint=None):
        url = self.package_url + '/{}/files/{}/download_token'.format(self.package_id, package_file_id)
        if custom_user_s3_endpoint:
            s3_dest_bucket, s3_dest_prefix = Utils.deconstruct_s3_url(custom_user_s3_endpoint)
            url += '?s3SourceBucket={}'.format(s3_dest_bucket)
            if s3_dest_prefix:
                url += '&s3SourcePrefix={}'.format(s3_dest_prefix)
        tmp = get_request(url, auth=self.auth,
                          error_handler=HttpErrorHandlingStrategy.reraise_status,
                          deserialize_handler=DeserializeHandler.none)
        return json.loads(tmp.text)

    def get_files_from_datastructure(self, data_structure):
        data_structure_regex = quote_plus('{}/.*'.format(data_structure))
        url = self.package_url + \
              '/{}/files?page=1&size=all&regex={}'.format(self.package_id, data_structure_regex)
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        results = json.loads(tmp.text)['results']
        return results

    def get_data_structure_manifest_file_info(self):
        url = self.package_url + \
              '/{}/files?page=1&size=all&types=Package%20Metadata&regex={}'.format(self.package_id,
                                                                                   'datastructure_manifest.txt')
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        results = json.loads(tmp.text)['results']
        # return None instead of empty list, since this method is always supposed to return 1 thing
        return results[0] if results else None

    def get_data_structure_files(self):
        url = self.package_url + \
              '/{}/files?page=1&size=all&types=Data'.format(self.package_id)
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        tmp.raise_for_status()
        results = json.loads(tmp.text)['results']
        return [r for r in results if r['nda_file_type'] == 'Data']

    def get_data_structure_file_info(self, short_name):
        url = self.package_url + \
              '/{}/files?page=1&size=all&types=Package%20Metadata&types=Data&regex={}'.format(self.package_id,
                                                                                              short_name)
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        tmp.raise_for_status()
        results = json.loads(tmp.text)['results']
        # return None instead of empty list, since this method is always supposed to return 1 thing
        return results[0] if results else None

    def get_package_file_info(self, file_id):
        url = self.package_url + '/{}/files/{}'.format(self.package_id, file_id)
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        return json.loads(tmp.text)

    def get_package_info(self):
        url = self.package_url + '/{}'.format(self.package_id)
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        return json.loads(tmp.text)

    def get_package_files_by_page(self, page, batch_size):
        url = self.package_url + '/{}/files?page={}&size={}'.format(self.package_id, page, batch_size)
        if self.regex_file_filter:
            url += '&regex={}'.format(self.regex_file_filter)
        try:
            tmp = get_request(url, auth=self.auth,
                              error_handler=HttpErrorHandlingStrategy.reraise_status, deserialize_handler=DeserializeHandler.none)
            tmp.raise_for_status()
            response = json.loads(tmp.text)
            return response['results']

        except HTTPError as e:
            if 'Cannot navigate past last page' in e.response.text:
                # empty alias and package-file return means there are no files for given page parameter
                return []
            else:
                raise e
        except Exception as e:
            raise e

    def get_presigned_urls(self, id_list):
        """
        Stores key-value pairs of (key: package_file_id, value: presigned URL)
        :param id_list: List of package file IDs with max size of 50,000
        """
        # Use the batchGeneratePresignedUrls when retrieving multiple files
        if not self.verify_flg:
            logger.debug('Retrieving credentials for {} files'.format(len(id_list)))
        url = self.package_url + '/{}/files/batchGeneratePresignedUrls'.format(self.package_id)
        tmp = post_request(url,  payload=id_list, auth=self.auth,
                           error_handler=HttpErrorHandlingStrategy.reraise_status, deserialize_handler=DeserializeHandler.none)
        response = json.loads(tmp.text)
        creds = {e['package_file_id']: e['downloadURL'] for e in response['presignedUrls']}
        if not self.verify_flg:
            logger.debug('Finished retrieving credentials')
        return creds

    def get_completed_files_in_download(self):
        download_progress_report_path = os.path.join(self.downloadcmd_package_metadata_directory,
                                                     '.download-progress', self.download_job_uuid,
                                                     'download-progress-report.csv')

        files = []
        if os.path.exists(download_progress_report_path):
            with open(download_progress_report_path, newline='') as csvfile:
                file_reader = csv.DictReader(csvfile)
                files = [f for f in file_reader if bool(f['exists'])]
        return files


import logging
import sys

import NDATools

from NDATools.clientscripts.vtcmd import configure, parse_args
from unittest import TestCase
try:
    import mock
except ImportError:
    from mock import patch


class TestConfiguration(TestCase):
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s:%(levelname)s:%(message)s")

    def setUp(self):
        self.test_file = r'test_files/image03_manifest/image03_w_populated_manifest_empty_image_file.csv'
        self.test_directory = r'test_files/image03_manifest'
        self.test_manifest_directory = r'test_files/image03_manifest'
        self.test_title = 'nda-tools-configuration-integration-test'
        self.test_description = 'nda-tools-configuration-integration-test'

    def test_manifest_skip_file_check(self):
        test_file = r'test_files/image03_manifest/image03_w_populated_manifest_empty_image_file_windows-path.csv'

        test_collection_id = 1860

        testargs = ['files', test_file, '-l', self.test_directory, '-m', self.test_manifest_directory, '-b', '-t',
                    self.test_title, '-d', self.test_description, '-c', str(test_collection_id),
                    '--skipLocalAssocFileCheck']

        with mock.patch.object(sys, 'argv', testargs):
            args = parse_args()
            config = configure(args)
            self._check_endpoints(config)
            self.assertEqual(
                args.files[0].encode('unicode-escape').decode(),
                test_file.encode('unicode-escape').decode())
            self.assertTrue(config.manifest_path)
            self.assertTrue(args.buildPackage)
            self.assertEqual(config.title, self.test_title)
            self.assertEqual(config.description, self.test_description)
            self.assertEqual(config.collection_id, test_collection_id)
            self.assertTrue(config.skip_local_file_check)

    def test_resume_submission_flags(self):
        test_file = r'test_files/image03_manifest/image03_w_populated_manifest_empty_image_file_windows-path.csv'
        worker_threads = 1

        testargs = ['files', test_file, '-r', '--hideProgress', '-wt', str(worker_threads)]

        with mock.patch.object(sys, 'argv', testargs):
            args = parse_args()
            config = configure(args)
            self._check_endpoints(config)
            self.assertEqual(
                args.files[0].encode('unicode-escape').decode(),
                test_file.encode('unicode-escape').decode())
            self.assertFalse(config.skip_local_file_check, config.skip_local_file_check)
            self.assertTrue(args.resume, args.resume)
            self.assertTrue(config.hideProgress, config.hideProgress)
            self.assertFalse(config.manifest_path, config.manifest_path)
            self.assertEqual(args.workerThreads, worker_threads)

    def test_csv_filelist(self):
        test_file = r'test_files/MultipleDataTypes/image03.csv'
        test_directory = r'test_files/MultipleDataTypes/image03'
        test_collection_id = 1860

        testargs = ['files', test_file, '-l', test_directory, '-b', '-t', self.test_title, '-d', self.test_description,
                    '-c', str(test_collection_id)]

        with mock.patch.object(sys, 'argv', testargs):
            args = parse_args()
            config = configure(args)
            self._check_endpoints(config)
            self.assertEqual(
                args.files[0].encode('unicode-escape').decode(),
                test_file.encode('unicode-escape').decode())
            self.assertTrue(args.buildPackage)
            self.assertEqual(config.title, self.test_title)
            self.assertEqual(config.description, self.test_description)
            self.assertEqual(config.collection_id, test_collection_id)
            self.assertFalse(config.manifest_path)

    def test_s3(self):
        test_file = r'test_files/MultipleDataTypes/image03_mixed_file_size_test_case_stage.csv'
        test_s3_bucket = 'NDAR_Central_1_dev'
        test_s3_bucket_prefix = 'submission_23596'
        test_collection_id = 17

        testargs = ['files', test_file, '-s3', test_s3_bucket, '-pre', test_s3_bucket_prefix, '-b',
                    '-c', str(test_collection_id)]

        with mock.patch.object(sys, 'argv', testargs):
            args = parse_args()
            config = configure(args)
            self._check_endpoints(config)
            self.assertEqual(
                args.files[0].encode('unicode-escape').decode(),
                test_file.encode('unicode-escape').decode())
            self.assertEqual(config.source_bucket, test_s3_bucket)
            self.assertEqual(config.source_prefix, test_s3_bucket_prefix)
            self.assertTrue(args.buildPackage)
            self.assertEqual(config.collection_id, test_collection_id)
            self.assertFalse(config.manifest_path)

    def test_s3_resume(self):
        test_submission_id = 23607
        test_s3_bucket = 'NDAR_Central_1_dev'
        test_s3_bucket_prefix = 'submission_23596'
        test_collection_id = 17

        testargs = ['files', '-r', str(test_submission_id), '-s3', test_s3_bucket, '-pre', test_s3_bucket_prefix, '-b',
                    '-c', str(test_collection_id)]

        with mock.patch.object(sys, 'argv', testargs):
            args = parse_args()
            config = configure(args)
            self._check_endpoints(config)
            self.assertTrue(args.resume, args.resume)
            self.assertTrue(args.files[0], test_submission_id)
            self.assertEqual(config.source_bucket, test_s3_bucket)
            self.assertEqual(config.source_prefix, test_s3_bucket_prefix)
            self.assertTrue(args.buildPackage)
            self.assertEqual(config.collection_id, test_collection_id)
            self.assertFalse(config.manifest_path, msg='Manifest')

    def _check_endpoints(self, config):
        logging.debug('Checking endpoint configuration...')
        self.assertTrue(config.datamanager_api is not None or config.datamanager_api != "", config.datamanager_api)
        self.assertTrue(config.validation_api is not None or config.validation_api != "", config.validation_api)
        self.assertTrue(config.submission_package_api is not None or config.submission_package_api != "", config.submission_package_api)
        self.assertTrue(config.submission_api is not None or config.submission_api != "", config.submission_api)
        self.assertTrue(config.validationtool_api is not None or config.validationtool_api != "", config.validationtool_api)
        self.assertTrue(config.validation_results is not None or config.validation_results != "", config.validation_results)
        self.assertTrue(config.submission_packages is not None or config.submission_packages != "", config.submission_packages)



from requests.adapters import HTTPAdapter
from urllib3.poolmanager import PoolManager

'''
This adapter is necessary for the download logic when files are located in alt endpoints and the presigned urls
generated from the query-package-service contain '.' characters
Reason -   VHostCalling format fails SSL certification when bucket contains '.'
see https://stackoverflow.com/questions/34154791/foo-bar-com-s3-amazonaws-com-doesnt-match-either-of-s3-amazonaws-com-s3
'''
class AltEndpointSSLAdapter(HTTPAdapter):

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(
            num_pools=connections, maxsize=maxsize,
            block=block, assert_hostname='s3.amazonaws.com')


from __future__ import absolute_import, print_function, with_statement

import sys

import NDATools

if sys.version_info[0] < 3:
    input = raw_input
import argparse
from NDATools.Download import Download
from NDATools.Configuration import *

logger = logging.getLogger(__name__)
def parse_args():
    parser = argparse.ArgumentParser(
        description='This application allows you to download files from an NDA package. Tutorials for creating packages'
                    ' can be found on the website (links provided below). Information for packages, including package-ids,'
                    ' are displayed on the packages dashboard page (https://nda.nih.gov/user/dashboard/packages.html). Users can'
                    ' only download data from "personal" type packages. To download files from a'
                    ' "shared" package you need to convert it to a "personal" package first, which can be done by clicking the "Add to my data'
                    ' packages" button in the actions dropdown. '
                    ''
                    '\nLinks:'
                    '\n\tvideo tutorial - https://nda.nih.gov/tutorials/nda/accessing_files_in_the_cloud.html?chapter=creating-a-package '
                    '\n\tpdf - https://ndar.nih.gov/ndarpublicweb/Documents/Accessing+Shared+Data+Sept_2021-1.pdf',
        formatter_class=argparse.RawTextHelpFormatter)

    optional = parser._action_groups.pop()
    required = parser.add_argument_group('required arguments')
    parser._action_groups.append(optional)

    required.add_argument('-dp', '--package', required=True, metavar='<package-id>', type=int, action='store',
                        help='The package-id containing the files you wish to download. If no other command-line '
                             'options are provided, the program will download all files from the specified package.')

    parser.add_argument('paths', metavar='<S3_path_list>', type=str, nargs='*', action='store',
                        help='Opional. When provided, the program will download only the specified files from the package.'
                             ' The specified files must exist in the package indicated by the -dp argument and the paths must be valid'
                             ' s3 urls.')

    parser.add_argument('-t', '--txt', metavar='<s3-links-file>', type=str, action='store',
                        help='Flags that a text file has been entered from where to download S3 files.')

    parser.add_argument('-ds', '--datastructure', metavar='<structure short-name>', type=str, action='store',
                        help='''Downloads all the files in a package from the specified data-structure. 
For example, to download all the image03 files from your package 12345, you should enter:
    downloadcmd -dp 12345 -ds image03
Note - the program only recognizes the short-names of the data-structures. The short-name is listed on the data-structures page 
and always ends in a 2 digit number. (For example, see the data-structure page for image03 at https://nda.nih.gov/data_structure.html?short_name=image03)''')

    parser.add_argument('-u', '--username', metavar='<username>', type=str, action='store',
                        help='NDA username')

    parser.add_argument('-p', '--password', help='Warning: Support for this setting has been deprecated and will no longer be '
                                                 'used by this tool. This option will be removed in future releases')

    parser.add_argument('-d', '--directory', metavar='<download_directory>', type=str, nargs=1, action='store',
                        help='Enter an alternate full directory path where you would like your files to be saved. The default is ~/NDA/nda-tools/<package-id>')

    parser.add_argument('-wt', '--workerThreads', metavar='<thread-count>', type=int, action='store',
                        help='''Specifies the number of downloads to attempt in parallel. For example, running 'downloadcmd -dp 12345 -wt 10' will 
cause the program to download a maximum of 10 files simultaneously until all of the files from package 12345 have been downloaded. 
A default value is calculated based on the number of cpus found on the machine, however a higher value can be chosen to decrease download times. 
If this value is set too high the download will slow. With 32 GB of RAM, a value of '10' is probably close to the maximum number of 
parallel downloads that the computer can handle''')

    parser.add_argument('--file-regex', metavar='<regular expression>',
                        help='''Option can be used to download only a subset of the files in a package.  This command line arg can be used with
the -ds, -dp or -t flags. 

Examples - 
1) To download all files with a ".txt" extension, you can use the regular expression .*.txt
    downloadcmd -dp 12345 --file-regex .*.txt
2) To download all files that contain "NDARINVZLHFUAF0" in the name, you can use the regular expression NDARINVZLHFUAF0 
    downloadcmd -dp 12345 -ds image03 --file-regex NDARINVZLHFUAF0  
3) Finally to download all files underneath a folder called "T1w" you can use the regular expression .*/T1w/.* 
    downloadcmd -dp 12345 -t s3-links.txt --file-regex .*/T1w/.*''')

    parser.add_argument('--verify', action='store_true',
                        help='''When this option is provided a download is not initiated. Instead, a csv file is produced that contains a record of 
the files in the download, along with information about the file-size if the file could be found on the computer. For large packages containing millions of files, 
this verification step can take hours (this can be even longer if files are stored on a network drive). When the program finishes, a few new files/folders 
will be created (if they don't already exist):
1) verification_report folder in the NDA/nda-tools/downloadcmd/packages/<package-id> directory
2) .download_progress folder (hidden) in the NDA/nda-tools/downloadcmd/packages/<package-id> directory, which is used to values between command invocations.
    a. .download_progress/download-job-manifest.csv file - contains entries mapping 
    b. UUID folders inside .download_progress (with names like '6a056ac4-2dd9-48f2-b921-44b29c883578')
3) download-verification-report.csv in the NDA/nda-tools/downloadcmd/packages/<package-id> directory
4) download-verification-retry-s3-links.csv in the NDA/nda-tools/downloadcmd/packages/<package-id> directory

The hidden folder listed in 2 contains special files used by the program to avoid re-running expensive, time-consuming processes. This folder should not be deleted.

The download-verification-report.csv file will contain a record for each file in the download and contain 6 columns :
1) 'package_file_id'
2) 'package_file_expected_location' - base path is the value provided for the -d/--directory arg
3) 'nda_s3_url'
4) 'exists' - value for column will be ('Y'/'N')
5) 'expected_size'
6) 'actual_file_size' - value for columnw will be '0' if file doesn't exist

In addition, the file will contain 1 header line which will provide the parameters used for the download (more information below). 

If this file is opened in Excel or Google Docs, the user can easily find information on specific files that they are interested in. 

This file can be useful but may contain more information than is needed. The download-verification-retry-s3-links.csv file contains the s3 links for all of the files 
in the download-verification-report.csv where EXISTS = 'N' or EXPECTED-SIZE does not equal ACTUAL-FILE-SIZE. If the user is only interested in re-initiating the download 
for the files that failed they can do so by using the  download-verification-retry-s3-links.csv as the value for the -t argument. i.e.

downloadcmd -dp <package-id> -t NDA/nda-tools/downloadcmd/packages/<package-id>/download-verification-retry-s3-links.csv 

When the --verify option is provided, the rest of the arguments provided to the command are used to determine what files are supposed to be included in the download. 

For example, if the user runs:  
   downloadcmd -dp 12345 --verify
The download-verification-report.csv file will contain a record for each file in the package 12345. Since no -d/--directory argument is provided, the program 
will check for the existance of the files in the default download location. 

If the user runs:
   downloadcmd -dp 12345 -d /home/myuser/customdirectory --verify
The download-verification-report.csv file will contain a record for each file in the package 12345 and will check for the existance of files in the /foo/bar

If the user runs:
   downloadcmd -dp 12345 -d /home/myuser/customdirectory -t file-with-s3-links.csv --verify
The download-verification-report.csv file will contain a record for each file listed in the file-with-s3-links.csv and will check for the existance of files in /foo/bar

If the user runs:
   downloadcmd -dp 12345 -d /home/myuser/customdirectory -ds image03 --verify
The download-verification-report.csv file will contain a record for each file in the package's image03 data-structure and will check for the existance of files in /foo/bar

If the user runs:
   downloadcmd -dp 12345 -d /home/myuser/customdirectory -ds image03 --file-regex --verify
The download-verification-report.csv file will contain a record for each file in the package's image03 data-structure which also matches the file-regex and will check 
for the existance of files in /foo/bar

NOTE - at the moment, this option cannot be used to verify downloads to s3 locations (see -s3 option below). That will be implemented in the near
future.''')

    parser.add_argument('-s3', '--s3-destination', metavar='<s3 bucket>',
                        help='''Specify s3 location which you would like to download your files to. When this option is specified, an attempt will be made
to copy the files from your package, which are stored in NDA's own S3 repository, to the S3 bucket provided. 

For s3-to-s3 copy operations to be successful, the s3 bucket supplied as the program argument must be configured to allow PUT object 
operations  for 'arn:aws:sts::618523879050:federated-user/<username>' where <username> is your nda username. 
For non-public buckets, this will require an update to your bucket's policy. The following statement should be sufficient to grant the uploading privileges necessary 
to run this program using the s3 argument after replacing <your-s3-bucket> with the bucket name:
         {
            "Sid": "AllowNDAUpload",
            "Effect": "Allow",
            "Principal": {
                "AWS": "arn:aws:iam::618523879050:federated-user/<username>"		
            },
            "Action": "s3:PutObject*",
            "Resource": "arn:aws:s3:::<your-s3-bucket>/*"
        }
        
You may need to email your company/institution IT department to have this added for you.
Note: If your bucket is encrypted with a customer-managed KMS key, then additional configuration is needed. 
For more details, check the information on the README page.
''')

    args = parser.parse_args()

    if args.password:
        print('Warning: Support for the password flag (-p, --password) has been removed from nda-tools due to security '
              'concerns and has been replaced with keyring.')
        args.__dict__.pop('password')

    return args


def configure(args):
    if os.path.isfile(os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg')):
        config = ClientConfiguration(os.path.join(os.path.expanduser('~'), '.NDATools/settings.cfg'), args.username)
        config.read_user_credentials()
    else:
        config = ClientConfiguration('clientscripts/config/settings.cfg', args.username)
        config.read_user_credentials()
        config.make_config()

    LoggingConfiguration.load_config(NDATools.NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER)

    return config


def main():
    args = parse_args()
    config = configure(args)

    if args.s3_destination and not args.s3_destination.startswith('s3://'):
        raise Exception(
            'Invalid argument for -s3 option :{}. Argument must start with "s3://"'.format(args.s3_destination))
    if sys.version_info < (3, 5):
        logger.error('ERROR: "--verify" only works with python 3.5 or later. Please upgrade Python in order to continue')
        exit_client()

    s3Download = Download(config, args)
    if args.verify:
        s3Download.verify_download()
    else:
        s3Download.start()


if __name__ == "__main__":
    main()


#!/usr/bin/env python

from setuptools import find_packages, setup

import NDATools

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
        name='nda_tools',
        description="NIMH Data Archive Python Client",
        install_requires=['boto3', 'botocore', 'tqdm', 'requests', 'mock', 'packaging','pyyaml', 'keyring'],
        extras_require={'test': ['pytest', 'pytest-datadir']},
        version= NDATools.__version__,
        long_description=long_description,
        long_description_content_type="text/markdown",
        author='NDA',
        author_email='NDAHelp@mail.nih.gov',
        url="https://github.com/NDAR/nda-tools/tree/master/NDATools",
        license='MIT',
        packages=find_packages(),
        include_package_data=True,
        data_files=[('config', ['NDATools/clientscripts/config/settings.cfg', 'NDATools/clientscripts/config/logging.yml'])],
        entry_points={
            'console_scripts': [
                'vtcmd = NDATools.clientscripts.vtcmd:main',
                'downloadcmd = NDATools.clientscripts.downloadcmd:main',
                'unit_tests = tests.run_unit_tests:main',
                'integration_tests = tests.run_integration_tests:main']
        }
    )
